import os
from typing import Any, Callable

from prefect.tasks.secrets import PrefectSecret
from prefect.utilities.tasks import defaults_from_attrs


class EnvVarSecret(PrefectSecret):
    """
    A `Secret` task that retrieves a value from an environment variable.

    Args:
        - name (str): the environment variable that contains the secret value
        - cast (Callable[[Any], Any]): A function that will be called on the Parameter
            value to coerce it to a type.
        - raise_if_missing (bool): if True, an error will be raised if the env var is not found.
        - **kwargs (Any, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        name: str,
        cast: Callable[[Any], Any] = None,
        raise_if_missing: bool = False,
        **kwargs
    ):
        self.cast = cast
        self.raise_if_missing = raise_if_missing
        super().__init__(name=name, **kwargs)

    @defaults_from_attrs("name")
    def run(self, name: str = None):
        """
        Returns the value of an environment variable after applying an optional `cast` function.

        Args:
            - name (str, optional): the name of the underlying environment variable to retrieve. Defaults
                to the name provided at initialization.

        Returns:
            - Any: the (optionally type-cast) value of the environment variable

        Raises:
            - ValueError: if `raise_is_missing` is `True` and the environment variable was not found
        """
        if self.raise_if_missing and name not in os.environ:
            raise ValueError("Environment variable not set: {}".format(name))
        value = os.getenv(name)
        if value is not None and self.cast is not None:
            value = self.cast(value)
        return value



import heapq
import itertools
import operator
import warnings
from datetime import datetime, timedelta
from typing import Callable, Iterable, List, Optional, cast

import prefect.schedules.adjustments
import prefect.schedules.clocks
import prefect.schedules.filters


class Schedule:
    """
    Schedules are used to generate dates for flow runs. Scheduling logic works as follows: First off, 
    candidate events are emitted by one or more `clocks`. Secondly, if filters were specified, 
    they are applied in this order: all `filters` must return True, at least one `or_filter` must 
    return True, then all `not_filters` must return False. Thridly, events that pass the filters 
    are adjusted based on the `adjustments` functions. Finally, the resulting `datetime` is emitted.

    Example:

    ```python
    from datetime import time, timedelta
    from prefect.schedules import Schedule, filters
    from prefect.schedules.clocks import IntervalClock

    schedule = Schedule(
        # emit an event every hour
        clocks=[IntervalClock(interval=timedelta(hours=1))],

        # only include weekdays
        filters=[filters.is_weekday],

        # only include 9am and 5pm
        or_filters=[filters.between_times(time(9), time(9)), filters.between_times(time(17), time(17))]
    )

    schedule.next(4) # returns the next 4 occurences of 9am and 5pm on weekdays
    ```

    Args:
        - clocks (List[prefect.schedules.clocks.Clock]): one or more clocks that emit events
            for this schedule. At least one clock is required.
        - filters (List[Callable[[datetime], bool]]): a list of filter functions that will be
            used to filter events. Events will only be emitted if all `filters` are True.
        - or_filters (List[Callable[[datetime], bool]]): a list of filter functions that
            will be used to filter events. Events will only be emitted if at least one of the
            `or_filters` are True
        - not_filters (List[Callable[[datetime], bool]]): a list of filter functions that
            will be used to filter events. Events will only be emitted if all `not_filters` are
            False
        - adjustments (List[Callable[[datetime], datetime]]): a list of adjustment functions
            that will be applied to dates that pass all filters. If more than one adjustment
            if provided, they will be applied in sequence.
    """

    def __init__(
        self,
        clocks: List["prefect.schedules.clocks.Clock"],
        filters: List[Callable[[datetime], bool]] = None,
        or_filters: List[Callable[[datetime], bool]] = None,
        not_filters: List[Callable[[datetime], bool]] = None,
        adjustments: List[Callable[[datetime], datetime]] = None,
    ):
        if not isinstance(clocks, Iterable):
            raise TypeError("clocks should be a list of clocks.")
        self.clocks = clocks
        self.filters = filters or []
        self.or_filters = or_filters or []
        self.not_filters = not_filters or []
        self.adjustments = adjustments or []

    @property
    def start_date(self) -> Optional[datetime]:
        return min([c.start_date for c in self.clocks if c.start_date], default=None)

    @property
    def end_date(self) -> Optional[datetime]:
        return max([c.end_date for c in self.clocks if c.end_date], default=None)

    def next(
        self, n: int, after: datetime = None, return_events: bool = False
    ) -> List[datetime]:
        """
        Retrieve the next `n` scheduled times, optionally after a specified date.

        Because clocks are potentially infinite, and filters may be prohibitive, this operation
        will stop after checking 10,000 events, no matter how many valid dates have been found.

        Args:
            - n (int): the number of dates to return
            - after (datetime): an optional starting point. All returned dates will be after this
                time.
            - return_events (bool, optional): an optional boolean specifying whether to return a full
                Clock Event or just the start_time of the associated event; defaults to `False`

        Returns:
            - List[datetime]: a list of datetimes
        """
        events = []
        counter = 0
        for event in self._get_clock_events(after=after):
            counter += 1
            if self._check_filters(event.start_time):
                event.start_time = self._apply_adjustments(event.start_time)
                events.append(event if return_events else event.start_time)
            if len(events) == n or counter >= 10000:
                break

        return events  # type: ignore

    def _get_clock_events(
        self, after: datetime = None
    ) -> Iterable["prefect.schedules.clocks.ClockEvent"]:
        """
        A generator of events emitted by the schedule's clocks.

        Events are sorted and unique (if two clocks emit the same date, it is only yielded once.)

        Args:
            - after (datetime): an optional starting point. All returned dates will be after this
                time.

        Returns:
            - Iterable[datetime]: An iterator of dates (as a generator)

        """
        clock_events = [clock.events(after=after) for clock in self.clocks]
        sorted_events = heapq.merge(*clock_events)

        # this next line yields items only if they differ from the previous item, which means
        # this generator only yields unique events (since the input is sorted)
        #
        # code from `unique_justseen()` at https://docs.python.org/3/library/itertools.html#itertools-recipes
        unique_events = map(
            next, map(operator.itemgetter(1), itertools.groupby(sorted_events))
        )  # type: Iterable[prefect.schedules.clocks.ClockEvent]
        yield from unique_events

    def _check_filters(self, dt: datetime) -> bool:
        """
        Check the schedule's filters:
            - all `filters` must pass
            - at least one of the `or_filters` must pass
            - none of the `not_filters` may pass

        Args:
            - dt (datetime): the date to check filters against

        Returns:
            - bool: True if the filters pass; False otherwise
        """
        # check that all `filters` pass
        all_filters = all(filter_fn(dt) for filter_fn in self.filters)
        # check that at least one `or_filter` passes
        or_filters = (
            any(filter_fn(dt) for filter_fn in self.or_filters)
            if self.or_filters
            else True
        )
        # check that no `not_filters` pass
        not_filters = (
            not any(filter_fn(dt) for filter_fn in self.not_filters)
            if self.not_filters
            else True
        )
        # return True if all three filter types passed
        return all_filters and or_filters and not_filters

    def _apply_adjustments(self, dt: datetime) -> datetime:
        """
        Apply the schedule's adjustments to a date.

        Args:
            - dt (datetime): the date to adjust

        Returns:
            - datetime: the adjusted date
        """
        # run date through adjustment pipeline
        for adjust_fn in self.adjustments:
            dt = adjust_fn(dt)
        return dt


# FIXME the proper signature for this function should be:
# interval (required), start_date (optional), end_date (optional)
# but start_date is currently first to maintain compatibility with an older version of
# Prefect
def IntervalSchedule(
    start_date: datetime = None, interval: timedelta = None, end_date: datetime = None
) -> Schedule:
    """
    A schedule formed by adding `timedelta` increments to a start_date.

    IntervalSchedules only support intervals of one minute or greater.

    NOTE: If the `IntervalSchedule` start time is provided with a DST-observing timezone,
    then the clock will adjust itself appropriately. Intervals greater than 24
    hours will follow DST conventions, while intervals of less than 24 hours will
    follow UTC intervals. For example, an hourly clock will fire every UTC hour,
    even across DST boundaries. When clocks are set back, this will result in two
    runs that *appear* to both be scheduled for 1am local time, even though they are
    an hour apart in UTC time. For longer intervals, like a daily clock, the
    interval clock will adjust for DST boundaries so that the clock-hour remains
    constant. This means that a daily clock that always fires at 9am will observe
    DST and continue to fire at 9am in the local time zone.

    Note that this behavior is different from the `CronSchedule`.

    Args:
        - interval (timedelta): interval on which this clock occurs
        - start_date (datetime, optional): first date of clock. If None, will be set to
            "2019-01-01 00:00:00 UTC"
        - end_date (datetime, optional): an optional end date for the clock

    Raises:
        - ValueError: if provided interval is less than one minute
    """
    return Schedule(
        clocks=[
            prefect.schedules.clocks.IntervalClock(
                interval=cast(timedelta, interval),  # due to FIXME, above
                start_date=start_date,
                end_date=end_date,
            )
        ]
    )


def CronSchedule(
    cron: str, start_date: datetime = None, end_date: datetime = None
) -> Schedule:
    """
    Cron clock.

    NOTE: If the `CronSchedule's` start time is provided with a DST-observing timezone,
    then the clock will adjust itself. Cron's rules for DST are based on clock times,
    not intervals. This means that an hourly cron clock will fire on every new clock
    hour, not every elapsed hour; for example, when clocks are set back this will result
    in a two-hour pause as the clock will fire *the first time* 1am is reached and
    *the first time* 2am is reached, 120 minutes later. Longer clocks, such as one
    that fires at 9am every morning, will automatically adjust for DST.

    Note that this behavior is different from the `IntervalSchedule`.

    Args:
        - cron (str): a valid cron string
        - start_date (datetime, optional): an optional start date for the clock
        - end_date (datetime, optional): an optional end date for the clock

    Raises:
        - ValueError: if the cron string is invalid
    """
    return Schedule(
        clocks=[
            prefect.schedules.clocks.CronClock(
                cron=cron, start_date=start_date, end_date=end_date
            )
        ]
    )


def OneTimeSchedule(start_date: datetime) -> Schedule:
    """
    A schedule corresponding to a single date

    NOTE: this function is deprecated and maintained only for backwards-compatibility.
    """
    warnings.warn(
        "The OneTimeSchedule is deprecated and will be removed from "
        "Prefect. Use a Schedule with a single-date DatesClock instead.",
        UserWarning,
    )
    return Schedule(clocks=[prefect.schedules.clocks.DatesClock(dates=[start_date])])


def UnionSchedule(schedules: List[Schedule]) -> Schedule:
    """
    A schedule formed by combining other schedules.

    NOTE: this function is deprecated and maintained only for backwards-compatibility.
    """
    warnings.warn(
        "The UnionSchedule is deprecated and will be removed from "
        "Prefect. Use a Schedule with multiple clocks instead.",
        UserWarning,
    )
    return Schedule(clocks=[c for s in schedules for c in s.clocks])



"""
Prefect Executors implement the logic for how Tasks are run. The standard interface
for an Executor consists of the following methods:

- `submit(fn, *args, **kwargs)`: submit `fn(*args, **kwargs)` for execution;
    note that this function is (in general) non-blocking, meaning that `executor.submit(...)`
    will _immediately_ return a future-like object regardless of whether `fn(*args, **kwargs)`
    has completed running
- `wait(object)`: resolves any objects returned by `executor.submit` to
    their values; this function _will_ block until execution of `object` is complete

Currently, the available executor options are:

- `LocalExecutor`: the no frills, straightforward executor - great for debugging;
    tasks are executed immediately upon being called by `executor.submit()`.Note
    that the `LocalExecutor` is not capable of parallelism.  Currently the default executor.
- `LocalDaskExecutor`: an executor that runs on `dask` primitives with a
    configurable dask scheduler.
- `DaskExecutor`: the most feature-rich of the executors, this executor runs
    on `dask.distributed` and has support for multiprocessing, multithreading, and distributed execution.

Which executor you choose depends on whether you intend to use things like parallelism
of task execution.

The key difference between the `LocalDaskExecutor` and the `DaskExecutor` is the choice
of scheduler. The `LocalDaskExecutor` is configurable to use
[any number of schedulers](https://docs.dask.org/en/latest/scheduler-overview.html) while the
`DaskExecutor` uses the [distributed scheduler](https://docs.dask.org/en/latest/scheduling.html).
This means that the `LocalDaskExecutor` can help achieve some multithreading / multiprocessing
however it does not provide as many distributed features as the `DaskExecutor`.

"""
import prefect
from prefect.engine.executors.base import Executor
from prefect.engine.executors.dask import DaskExecutor, LocalDaskExecutor
from prefect.engine.executors.local import LocalExecutor



"""
A collection of tasks for interacting with Airtable.
"""
try:
    from prefect.tasks.airtable.airtable import WriteAirtableRow
except ImportError:
    raise ImportError(
        'Using `prefect.tasks.airtable` requires Prefect to be installed with the "airtable" extra.'
    )



from typing import Any

import prefect
from prefect import Task, Flow

from .conditional import CompareValue


__all__ = ("case",)


class case(object):
    """A conditional block in a flow definition.

    Used as a context-manager, ``case`` creates a block of tasks that are only
    run if the result of ``task`` is equal to ``value``.

    Args:
        - task (Task): The task to use in the comparison
        - value (Any): A constant the result of ``task`` will be compared with

    Example:

    A ``case`` block is similar to Python's if-blocks. It delimits a block
    of tasks that will only be run if the result of ``task`` is equal to
    ``value``:

    ```python
    # Standard python code
    if task == value:
        res = run_if_task_equals_value()
        other_task(res)

    # Equivalent prefect code
    with case(task, value):
        # Tasks created in this block are only run if the
        # result of ``task`` is equal to ``value``
        res = run_if_task_equals_value()
        other_task(run)
    ```

    The ``value`` argument can be any non-task object. Here we branch on a
    string result:

    ```python
    with Flow("example") as flow:
        cond = condition()

        with case(cond, "a"):
            run_if_cond_is_a()

        with case(cond, "b"):
            run_if_cond_is_b()
    ```
    """

    def __init__(self, task: Task, value: Any):
        if isinstance(value, Task):
            raise TypeError("`value` cannot be a task")

        self.task = task
        self.value = value
        self._tasks = set()
        self._flow = None

    def add_task(self, task: Task, flow: Flow) -> None:
        """Add a new task under the case statement.

        Args:
            - task (Task): the task to add
            - flow (Flow): the flow to use
        """
        if self._flow is None:
            self._flow = flow
        elif self._flow is not flow:
            raise ValueError(
                "Multiple flows cannot be used with the same case statement"
            )
        self._tasks.add(task)

    def __enter__(self):
        self.__prev_case = prefect.context.get("case")
        prefect.context.update(case=self)

    def __exit__(self, *args):
        if self.__prev_case is None:
            prefect.context.pop("case", None)
        else:
            prefect.context.update(case=self.__prev_case)

        if self._tasks:
            cond = CompareValue(self.value, name=f"case({self.value})",).bind(
                value=self.task, flow=self._flow
            )

            for child in self._tasks:
                # If a task has no upstream tasks created in this case block,
                # the case conditional should be set as an upstream task.
                if not self._tasks.intersection(self._flow.upstream_tasks(child)):
                    child.set_upstream(cond, flow=self._flow)



import os
import tempfile
from subprocess import PIPE, STDOUT, Popen
from typing import Any

import prefect
from prefect.utilities.tasks import defaults_from_attrs


class ShellTask(prefect.Task):
    """
    Task for running arbitrary shell commands.

    Args:
        - command (string, optional): shell command to be executed; can also be
            provided post-initialization by calling this task instance
        - env (dict, optional): dictionary of environment variables to use for
            the subprocess; can also be provided at runtime
        - helper_script (str, optional): a string representing a shell script, which
            will be executed prior to the `command` in the same process. Can be used to
            change directories, define helper functions, etc. when re-using this Task
            for different commands in a Flow
        - shell (string, optional): shell to run the command with; defaults to "bash"
        - return_all (bool, optional): boolean specifying whether this task should return all lines of stdout
            as a list, or just the last line as a string; defaults to `False`
        - **kwargs: additional keyword arguments to pass to the Task constructor

    Example:
        ```python
        from prefect import Flow
        from prefect.tasks.shell import ShellTask

        task = ShellTask(helper_script="cd ~")
        with Flow("My Flow") as f:
            # both tasks will be executed in home directory
            contents = task(command='ls')
            mv_file = task(command='mv .vimrc /.vimrc')

        out = f.run()
        ```
    """

    def __init__(
        self,
        command: str = None,
        env: dict = None,
        helper_script: str = None,
        shell: str = "bash",
        return_all: bool = False,
        **kwargs: Any
    ):
        self.command = command
        self.env = env
        self.helper_script = helper_script
        self.shell = shell
        self.return_all = return_all
        super().__init__(**kwargs)

    @defaults_from_attrs("command", "env")
    def run(self, command: str = None, env: dict = None) -> str:
        """
        Run the shell command.

        Args:
            - command (string): shell command to be executed; can also be
                provided at task initialization. Any variables / functions defined in
                `self.helper_script` will be available in the same process this command
                runs in
            - env (dict, optional): dictionary of environment variables to use for
                the subprocess

        Returns:
            - stdout (string): if `return_all` is `False` (the default), only the last line of stdout
                is returned, otherwise all lines are returned, which is useful for passing
                result of shell command to other downstream tasks. If there is no output, `None` is returned.

        Raises:
            - prefect.engine.signals.FAIL: if command has an exit code other
                than 0
        """
        if command is None:
            raise TypeError("run() missing required argument: 'command'")

        current_env = os.environ.copy()
        current_env.update(env or {})
        with tempfile.NamedTemporaryFile(prefix="prefect-") as tmp:
            if self.helper_script:
                tmp.write(self.helper_script.encode())
                tmp.write("\n".encode())
            tmp.write(command.encode())
            tmp.flush()
            sub_process = Popen(
                [self.shell, tmp.name], stdout=PIPE, stderr=STDOUT, env=current_env
            )
            lines = []
            line = None
            for raw_line in iter(sub_process.stdout.readline, b""):
                line = raw_line.decode("utf-8").rstrip()
                if self.return_all:
                    lines.append(line)
                else:
                    # if we're returning all, we don't log every line
                    self.logger.debug(line)
            sub_process.wait()
            if sub_process.returncode:
                msg = "Command failed with exit code {0}: {1}".format(
                    sub_process.returncode, line
                )
                self.logger.error(msg)
                raise prefect.engine.signals.FAIL(msg) from None  # type: ignore
        if self.return_all:
            return lines
        else:
            return line



from typing import Any, List

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class GetRepoInfo(Task):
    """
    Task for retrieving GitHub repository information using the v3 version of the GitHub REST API.

    Args:
        - repo (str, optional): the name of the repository to open the issue in; must be provided in the
            form `organization/repo_name` or `user/repo_name`; can also be provided to the `run` method
        - info_keys (List[str], optional): a list of repo attributes to pull (e.g., `["stargazers_count", "subscribers_count"]`).
            A full list of available keys can be found in the official [GitHub documentation](https://developer.github.com/v3/repos/)
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task init method
    """

    def __init__(self, repo: str = None, info_keys: List[str] = None, **kwargs: Any):
        self.repo = repo
        self.info_keys = info_keys or []

        super().__init__(**kwargs)

    @defaults_from_attrs("repo", "info_keys")
    def run(
        self, repo: str = None, info_keys: List[str] = None, token: str = None
    ) -> None:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a Flow context,
        or by using `Task.bind`.

        Args:
            - repo (str, optional): the name of the repository to open the issue in; must be provided in the
                form `organization/repo_name`; defaults to the one provided at initialization
            - info_keys (List[str], optional): a list of repo attributes to pull (e.g., `["stargazers_count", "subscribers_count"]`).
                A full list of available keys can be found in the official [GitHub documentation](https://developer.github.com/v3/repos/)
            - token (str): a GitHub access token

        Raises:
            - ValueError: if a `repo` was never provided
            - HTTPError: if the GET request returns a non-200 status code

        Returns:
            - dict: dictionary of the requested information
        """
        if repo is None:
            raise ValueError("A GitHub repository must be provided.")

        # 'import requests' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import requests

        url = "https://api.github.com/repos/{}".format(repo)
        headers = {
            "AUTHORIZATION": "token {}".format(token),
            "Accept": "application/vnd.github.v3+json",
        }

        ## send the request
        resp = requests.get(url, headers=headers)
        resp.raise_for_status()
        data = resp.json()

        return {key: data[key] for key in info_keys}


class CreateBranch(Task):
    """
    Task for creating new branches using the v3 version of the GitHub REST API.

    Args:
        - repo (str, optional): the name of the repository to create the branch in; must be provided in the
            form `organization/repo_name` or `user/repo_name`; can also be provided to the `run` method
        - base (str, optional): the name of the branch you want to branch off; can also
            be provided to the `run` method.  Defaults to "master".
        - branch_name (str, optional): the name of the new branch; can also be provided to the `run` method
        - **kwargs (Any, optional): additional keyword arguments to pass to the standard Task init method
    """

    def __init__(
        self,
        repo: str = None,
        base: str = "master",
        branch_name: str = None,
        **kwargs: Any
    ):
        self.repo = repo
        self.base = base
        self.branch_name = branch_name

        super().__init__(**kwargs)

    @defaults_from_attrs("repo", "base", "branch_name")
    def run(
        self,
        repo: str = None,
        base: str = None,
        branch_name: str = None,
        token: str = None,
    ) -> dict:
        """
        Run method for this Task. Invoked by calling this Task after initialization within a Flow context,
        or by using `Task.bind`.

        Args:
            - repo (str, optional): the name of the repository to open the issue in; must be provided in the
                form `organization/repo_name`; defaults to the one provided at initialization
            - base (str, optional): the name of the branch you want to branch off; if not provided here,
                defaults to the one set at initialization
            - branch_name (str, optional): the name of the new branch; if not provided here, defaults to
                the one set at initialization
            - token (str): a GitHub access token

        Raises:
            - ValueError: if a `repo` or `branch_name` was never provided, or if the base branch wasn't found
            - HTTPError: if the GET request returns a non-200 status code

        Returns:
            - dict: dictionary of the response (includes commit hash, etc.)
        """
        if branch_name is None:
            raise ValueError("A branch name must be provided.")

        if repo is None:
            raise ValueError("A GitHub repository must be provided.")

        # 'import requests' is expensive time-wise, we should do this just-in-time to keep
        # the 'import prefect' time low
        import requests

        url = "https://api.github.com/repos/{}/git/refs".format(repo)
        headers = {
            "AUTHORIZATION": "token {}".format(token),
            "Accept": "application/vnd.github.v3+json",
        }

        ## gather branch information
        resp = requests.get(url + "/heads", headers=headers)
        resp.raise_for_status()
        branch_data = resp.json()

        commit_sha = None
        for branch in branch_data:
            if branch.get("ref") == "refs/heads/{}".format(base):
                commit_sha = branch.get("object", {}).get("sha")
                break

        if commit_sha is None:
            raise ValueError("Base branch {} not found.".format(base))

        ## create new branch
        new_branch = {"ref": "refs/heads/{}".format(branch_name), "sha": commit_sha}
        resp = requests.post(url, headers=headers, json=new_branch)
        resp.raise_for_status()
        return resp.json()



"""
Utility functions for interacting with Kubernetes API.
"""
from typing import Union

from kubernetes import client, config
from kubernetes.config.config_exception import ConfigException

from prefect.client import Secret


K8S_CLIENTS = {
    "job": client.BatchV1Api,
    "pod": client.CoreV1Api,
    "service": client.CoreV1Api,
    "deployment": client.ExtensionsV1beta1Api,
}


KubernetesClient = Union[
    client.BatchV1Api, client.CoreV1Api, client.ExtensionsV1beta1Api
]


def get_kubernetes_client(
    resource: str, kubernetes_api_key_secret: str = "KUBERNETES_API_KEY"
) -> KubernetesClient:
    """
    Utility function for loading kubernetes client object for a given resource.

    It will attempt to connect to a Kubernetes cluster in three steps with
    the first successful connection attempt becoming the mode of communication with a
    cluster.

    1. Attempt to use a Prefect Secret that contains a Kubernetes API Key. If
    `kubernetes_api_key_secret` = `None` then it will attempt the next two connection
    methods. By default the value is `KUBERNETES_API_KEY` so providing `None` acts as
    an override for the remote connection.
    2. Attempt in-cluster connection (will only work when running on a Pod in a cluster)
    3. Attempt out-of-cluster connection using the default location for a kube config file

    Args:
        - resource (str): the name of the resource to retrieve a client for. Currently
            you can use one of these values: `job`, `pod`, `service`, `deployment`
        - kubernetes_api_key_secret (str, optional): the name of the Prefect Secret
            which stored your Kubernetes API Key; this Secret must be a string and in
            BearerToken format

    Returns:
        - KubernetesClient: an initialized and authenticated Kubernetes Client
    """
    k8s_client = K8S_CLIENTS[resource]

    kubernetes_api_key = None
    if kubernetes_api_key_secret:
        kubernetes_api_key = Secret(kubernetes_api_key_secret).get()

    if kubernetes_api_key:
        configuration = client.Configuration()
        configuration.api_key["authorization"] = kubernetes_api_key
        k8s_client = k8s_client(client.ApiClient(configuration))
    else:
        try:
            config.load_incluster_config()
        except ConfigException:
            config.load_kube_config()

        k8s_client = k8s_client()

    return k8s_client



from typing import Any
from prefect.engine.result import Result


class ConstantResult(Result):
    """
    Hook for storing and retrieving constant Python objects. Only intended to be used
    internally.  The "backend" in this instance is the class instance itself.

    Args:
        - **kwargs (Any, optional): any additional `Result` initialization options
    """

    def __init__(self, **kwargs: Any) -> None:
        if "serializer" in kwargs:
            raise ValueError("Can't pass a serializer to a ConstantResult.")
        super().__init__(**kwargs)

    def read(self, location: str) -> Result:
        """
        Will return the underlying value regardless of the argument passed.

        Args:
            - location (str): an unused argument
        """
        return self

    def write(self, value: Any, **kwargs: Any) -> Result:
        """
        Will return the repr of the underlying value, purely for convenience.

        Args:
            - value (Any): unused, for interface compatibility
            - **kwargs (optional): unused, for interface compatibility

        Raises:
            - ValueError: ConstantResults cannot be written to
        """
        raise ValueError("Cannot write values to `ConstantResult` types.")

    def exists(self, location: str, **kwargs: Any) -> bool:
        """
        As all Python objects are valid constants, always returns `True`.

        Args:
             - location (str): for interface compatibility
             - **kwargs (Any): string format arguments for `location`

        Returns:
            - bool: True, confirming the constant exists.
        """
        return True



import redis

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class RedisSet(Task):
    """
    Task for setting a Redis key-value pair.

    Args:
        - host (str, optional): name of Redis host, defaults to 'localhost'
        - port (int, optional): Redis port, defaults to 6379
        - db (int, optional): redis database index, defaults to 0
        - password_secret (str, optional): the name of the Prefect Secret
            that stores your Redis credentials
        - redis_key (str, optional): Redis key to be set, can be provided at initialization or runtime
        - redis_val (Redis native type, optional): Redis val to be set, can be provided at initialization or runtime,
            Redis native types include strings, dictionaries, lists, sets, and sorted sets
        - redis_connection_params (dict, optional): key-value pairs passed to the redis.Redis connection
            initializer
        - ex (int, optional): if provided, sets an expire flag, in seconds, on 'redis_key' set
        - px (int, optional): if provided, sets an expire flag, in milliseconds, on 'redis_key' set
        - nx (int, optional): if set to True, set the value at 'redis_key' to 'redis_val' only
            if it does not exist, defaults to False
        - xx (int, optional): if set to True, set the value at 'redis_key' to 'redis_val' only
            if it already exists, defaults to False
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password_secret: str = "REDIS_PASSWORD",
        redis_key: str = None,
        redis_val=None,
        redis_connection_params: dict = None,
        ex: int = None,
        px: int = None,
        nx: bool = False,
        xx: bool = False,
        **kwargs
    ):
        self.host = host
        self.port = port
        self.db = db
        self.password_secret = password_secret
        self.redis_connection_params = redis_connection_params or {}

        self.redis_key = redis_key
        self.redis_val = redis_val
        self.ex = ex
        self.px = px
        self.nx = nx
        self.xx = xx
        super().__init__(**kwargs)

    @defaults_from_attrs("redis_key", "redis_val", "ex", "px", "nx", "xx")
    def run(
        self,
        redis_key: str = None,
        redis_val=None,
        ex: int = None,
        px: int = None,
        nx: bool = False,
        xx: bool = False,
    ):
        """
        Task run method. Sets Redis key-value pair.

        Args:
            - redis_key (str, optional): Redis key to be set, can be provided at initialization or runtime
            - redis_val (Redis native type, optional): Redis val to be set, can be provided at initialization or runtime,
                Redis native types include strings, dictionaries, lists, sets, and sorted sets
            - ex (int, optional): if provided, sets an expire flag, in seconds, on 'redis_key' set
            - px (int, optional): if provided, sets an expire flag, in milliseconds, on 'redis_key' set
            - nx (int, optional): if set to True, set the value at 'redis_key' to 'redis_val' only
                if it does not exist, defaults to False
            - xx (int, optional): if set to True, set the value at 'redis_key' to 'redis_val' only
                if it already exists, defaults to False

        Returns:
            - bool: status of set operation

        Raises:
            - ValueError: if redis_key or redis_val is not provided
        """
        if None in (redis_key, redis_val):
            raise ValueError("redis_key and redis_val must be provided")

        ## connect to redis
        password = Secret(self.password_secret).get()
        connection = redis.Redis(
            host=self.host,
            port=self.port,
            db=self.db,
            password=password,
            **self.redis_connection_params
        )

        result = connection.set(
            name=redis_key, value=redis_val, ex=ex, px=px, nx=nx, xx=xx
        )

        return result


class RedisGet(Task):
    """
    Task for getting a value based on key from a Redis connection.

    Args:
        - host (str, optional): name of Redis host, defaults to 'localhost'
        - port (int, optional): Redis port, defaults to 6379
        - db (int, optional): redis database index, defaults to 0
        - password_secret (str, optional): the name of the Prefect Secret
            that stores your Redis password
        - redis_connection_params (dict, optional): key-value pairs passed to the redis.Redis connection
            initializer
        - redis_key (str, optional): Redis key to get value, can be provided at initialization or runtime
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password_secret: str = "REDIS_PASSWORD",
        redis_connection_params: dict = None,
        redis_key: str = None,
        **kwargs
    ):
        self.host = host
        self.port = port
        self.db = db
        self.password_secret = password_secret
        self.redis_connection_params = redis_connection_params or {}

        self.redis_key = redis_key
        super().__init__(**kwargs)

    @defaults_from_attrs("redis_key")
    def run(self, redis_key: str = None):
        """
        Task run method.

        Args:
            - redis_key (str, optional): Redis key to get value, can be provided at initialization or runtime

        Returns:
            - value: value associated with redis_key

        Raises:
            - ValueError: if redis_key is not provided
        """
        if not redis_key:
            raise ValueError("redis_key must be provided")

        ## connect to redis
        password = Secret(self.password_secret).get()
        connection = redis.Redis(
            host=self.host,
            port=self.port,
            db=self.db,
            password=password,
            **self.redis_connection_params
        )

        result = connection.get(name=redis_key)

        return result


class RedisExecute(Task):
    """
    Task for executing a command against a Redis connection

    Args:
        - host (str, optional): name of Redis host, defaults to 'localhost'
        - port (int, optional): Redis port, defaults to 6379
        - db (int, optional): redis database index, defaults to 0
        - password_secret (str, optional): the name of the Prefect Secret
            that stores your Redis credentials
        - redis_connection_params (dict, optional): key-value pairs passed to the redis.Redis connection
            initializer
        - redis_cmd (str, optional): Redis command to execute, must be provided at initialization or
            runtime
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password_secret: str = "REDIS_PASSWORD",
        redis_connection_params: dict = None,
        redis_cmd: str = None,
        **kwargs
    ):
        self.host = host
        self.port = port
        self.db = db
        self.password_secret = password_secret
        self.redis_connection_params = redis_connection_params or {}

        self.redis_cmd = redis_cmd
        super().__init__(**kwargs)

    @defaults_from_attrs("redis_cmd")
    def run(self, redis_cmd: str = None):
        """
        Task run method. Executes a command against a Redis connection.

        Args:
            - redis_cmd (str, optional): Redis command to execute, must be provided at initialization or
                runtime

        Returns:
            - result: result of executed Redis command
        """
        if not redis_cmd:
            raise ValueError("A redis command must be specified")

        ## connect to redis
        password = Secret(self.password_secret).get()
        connection = redis.Redis(
            host=self.host,
            port=self.port,
            db=self.db,
            password=password,
            **self.redis_connection_params
        )

        result = connection.execute_command(redis_cmd)

        return result



"""
This module ensures that old-style schedules (pre-0.6.1) can be deserialized as new-style
schedules. It is deprecated.
"""
from typing import Any

from marshmallow import fields, post_load

import prefect
from prefect.utilities.serialization import (
    DateTimeTZ,
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
)


class IntervalScheduleSchema(ObjectSchema):
    class Meta:
        object_class = lambda: prefect.schedules.IntervalSchedule

    start_date = DateTimeTZ(required=True)
    end_date = DateTimeTZ(allow_none=True)
    interval = fields.TimeDelta(precision="microseconds", required=True)


class CronScheduleSchema(ObjectSchema):
    class Meta:
        object_class = lambda: prefect.schedules.CronSchedule

    start_date = DateTimeTZ(allow_none=True)
    end_date = DateTimeTZ(allow_none=True)
    cron = fields.String(required=True)


class OneTimeScheduleSchema(ObjectSchema):
    class Meta:
        object_class = lambda: prefect.schedules.OneTimeSchedule

    start_date = DateTimeTZ(required=True)


class UnionScheduleSchema(ObjectSchema):
    class Meta:
        object_class = lambda: prefect.schedules.UnionSchedule

    start_date = DateTimeTZ(required=True)
    end_date = DateTimeTZ(allow_none=True)
    schedules = fields.Nested(
        "prefect.serialization.schedule.ScheduleSchema", many=True
    )

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> prefect.schedules.Schedule:
        return super().create_object({"schedules": data.pop("schedules", [])})



import warnings

from prefect.client.secrets import Secret as _Secret
from prefect.core.task import Task
from prefect.engine.results import SecretResult
from prefect.utilities.tasks import defaults_from_attrs


class SecretBase(Task):
    """
    Base Secrets Task.  This task does not perform any action but rather serves as the base
    task class which should be inherited from when writing new Secret Tasks.

    Users should subclass this Task and override its `run` method for plugging into other Secret stores,
    as it is handled differently during execution to ensure the underlying secret value is not accidentally
    persisted in a non-safe location.

    Args:
        - **kwargs (Any, optional): additional keyword arguments to pass to the Task constructor

    Raises:
        - ValueError: if a `result` keyword is passed
    """

    def __init__(self, **kwargs):
        if kwargs.get("result"):
            raise ValueError("Result types for Secrets are not configurable.")
        kwargs["checkpoint"] = False
        super().__init__(**kwargs)
        self.result = SecretResult(secret_task=self)


class PrefectSecret(SecretBase):
    """
    Prefect Secrets Task.  This task retrieves the underlying secret through
    the Prefect Secrets API (which has the ability to toggle between local vs. Cloud secrets).

    Args:
        - name (str): The name of the underlying secret
        - **kwargs (Any, optional): additional keyword arguments to pass to the Task constructor

    Raises:
        - ValueError: if a `result` keyword is passed
    """

    def __init__(self, name, **kwargs):
        kwargs["name"] = name
        super().__init__(**kwargs)

    @defaults_from_attrs("name")
    def run(self, name: str = None):
        """
        The run method for Secret Tasks.  This method actually retrieves and returns the underlying secret value
        using the `Secret.get()` method.  Note that this method first checks context for the secret value, and if not
        found either raises an error or queries Prefect Cloud, depending on whether `config.cloud.use_local_secrets`
        is `True` or `False`.

        Args:
            - name (str, optional): the name of the underlying Secret to retrieve. Defaults
                to the name provided at initialization.

        Returns:
            - Any: the underlying value of the Prefect Secret
        """
        return _Secret(name).get()


class Secret(PrefectSecret):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "The `Secret` task is deprecated and has been renamed `PrefectSecret`.",
            UserWarning,
        )
        super().__init__(*args, **kwargs)



import keyword
from typing import Set

import prefect
from prefect.core.task import Task


def is_valid_identifier(string: str) -> bool:
    """
    Determines whether a string is a valid Python identifier (meaning it can
    be used as a variable name or keyword argument).

    Args:
        - string (str): string to be validated

    Returns:
        - bool: boolean specifying whether the provided string is a valid Python
            identifier

    Example:
        ```python
        >>> is_valid_identifier('hi5')
        True
        >>> is_valid_identifier('5hi')
        False
        >>> is_valid_identifier('from')
        False
        >>> is_valid_identifier('hi.5')
        False
        ```
    """
    return string.isidentifier() and not keyword.iskeyword(string)


class Edge:
    """
    Edges represent connections between Tasks.

    At a minimum, an edge links an `upstream_task` and a `downstream_task`
    indicating that the downstream task shouldn't attempt to run until the upstream
    task is complete.

    In addition, edges can specify a key that describe how upstream results
    are passed to the downstream task.

    Args:
        - upstream_task (Task): the task that must run before the `downstream_task`
        - downstream_task (Task): the task that will be run after the
            `upstream_task`. The upstream task state is passed to the
            downstream task's trigger function to determine whether the
            downstream task should run.
        - key (str, optional): Passing a key indicates
            that the upstream result should be passed to the downstream
            task as a keyword argument given by `key`.
        - mapped (bool, optional): boolean indicating whether this edge
            represents a mapped task; defaults to `False`

    The key indicates that the result of the upstream task should be passed
    to the downstream task under the key.

    In general, Edges are created and handled in the background by the [Flow](flow.html)
    class and will not be directly instantiated by users.

    Example:
        ```python
        from prefect import *
        from prefect.core import Edge

        class Add(Task):
            def run(self, x):
                return x + 1

        class Number(Task):
            def run(self):
                return 2

        # passes the result of the Number() task to Add() as 'x'
        edge = Edge(Number(), Add(), key='x')
        ```
    """

    def __init__(
        self,
        upstream_task: Task,
        downstream_task: Task,
        key: str = None,
        mapped: bool = False,
    ):
        if upstream_task is downstream_task:
            raise ValueError("Edges can not connect a task to itself.")

        self.upstream_task = upstream_task
        self.downstream_task = downstream_task
        self.mapped = mapped

        if key is not None:
            if not (isinstance(key, str) and is_valid_identifier(key)):
                raise ValueError(
                    'Key must be a valid identifier (received "{}")'.format(key)
                )
        self.key = key

    # Comparison --------------------------------------------------------------

    @property
    def tasks(self) -> Set[Task]:
        """
        Returns a set of both upstream and downstream tasks for this edge.
        """
        return {self.upstream_task, self.downstream_task}

    def __repr__(self) -> str:
        return "<Edge{k}: {u} to {d}>".format(
            u=self.upstream_task.name,
            d=self.downstream_task.name,
            k=" (key={})".format(self.key) if self.key else "",
        )

    def __eq__(self, other: "Edge") -> bool:  # type: ignore
        if type(self) == type(other):
            attrs = ["upstream_task", "downstream_task", "key", "mapped"]
            return all(getattr(self, a) == getattr(other, a) for a in attrs)
        return False

    def __hash__(self) -> int:
        return hash((self.upstream_task, self.downstream_task, self.key))

    def serialize(self) -> dict:
        """
        Represents the Edge as a dict.
        """
        return prefect.serialization.edge.EdgeSchema().dump(self)



"""
Result Handlers provide the hooks that Prefect uses to store task results in production; a `ResultHandler` can be provided to a `Flow` at creation.

Anytime a task needs its output or inputs stored, a result handler is used to determine where this data should be stored (and how it can be retrieved).
"""
import os
from typing import Any

import cloudpickle
import pendulum
from slugify import slugify

import prefect
from prefect.engine.result_handlers import ResultHandler


class LocalResultHandler(ResultHandler):
    """
    Hook for storing and retrieving task results from local file storage.
    Task results are written using `cloudpickle` and stored in the
    provided location for use in future runs.

    Args:
        - dir (str, optional): the _absolute_ path to a directory for storing
            all results; defaults to `${prefect.config.home_dir}/results`
        - validate (bool, optional): a boolean specifying whether to validate the
            provided directory path; if `True`, the directory will be converted to an
            absolute path and created.  Defaults to `True`
    """

    def __init__(self, dir: str = None, validate: bool = True):
        full_prefect_path = os.path.abspath(prefect.config.home_dir)
        if (
            dir is None
            or os.path.commonpath([full_prefect_path, os.path.abspath(dir)])
            == full_prefect_path
        ):
            directory = os.path.join(prefect.config.home_dir, "results")
        else:
            directory = dir

        if validate:
            abs_directory = os.path.abspath(os.path.expanduser(directory))
            if not os.path.exists(abs_directory):
                os.makedirs(abs_directory)
        else:
            abs_directory = directory
        self.dir = abs_directory
        super().__init__()

    def read(self, fpath: str) -> Any:
        """
        Read a result from the given file location.

        Args:
            - fpath (str): the _absolute_ path to the location of a written result

        Returns:
            - the read result from the provided file
        """
        self.logger.debug("Starting to read result from {}...".format(fpath))
        with open(fpath, "rb") as f:
            val = cloudpickle.loads(f.read())
        self.logger.debug("Finished reading result from {}...".format(fpath))
        return val

    def write(self, result: Any) -> str:
        """
        Serialize the provided result to local disk.

        Args:
            - result (Any): the result to write and store

        Returns:
            - str: the _absolute_ path to the written result on disk
        """
        fname = "prefect-result-" + slugify(pendulum.now("utc").isoformat())
        loc = os.path.join(self.dir, fname)
        self.logger.debug("Starting to upload result to {}...".format(loc))
        with open(loc, "wb") as f:
            f.write(cloudpickle.dumps(result))
        self.logger.debug("Finished uploading result to {}...".format(loc))
        return loc



import os
from typing import Dict, List, Union

import azureml.core.datastore
from azureml.core.workspace import Workspace
from azureml.data.azure_storage_datastore import (
    AbstractAzureStorageDatastore,
    AzureBlobDatastore,
)
from azureml.data.data_reference import DataReference

from prefect import Task
from prefect.client import Secret
from prefect.utilities.tasks import defaults_from_attrs


class DatastoreRegisterBlobContainer(Task):
    """
    Task for registering Azure Blob Storage container as a Datastore in a Azure ML service Workspace. 

    Args:
        - workspace (azureml.core.workspace.Workspace): The Workspace to which the Datastore is to be registered.
        - container_name (str, optional): The name of the container.
        - datastore_name (str, optional): The name of the datastore. If not defined, the container name will be used.
        - create_container_if_not_exists (bool, optional): Create a container, if one does not exist with the given name. 
        - overwrite_existing_datastore (bool, optional): Overwrite an existing datastore. If the datastore does not exist, it will be created.
        - azure_credentials_secret (str, optinonal): The name of the Prefect Secret that stores your Azure credentials; this Secret must be a JSON string with two keys: `ACCOUNT_NAME` and either `ACCOUNT_KEY` or `SAS_TOKEN` (if both are defined then`ACCOUNT_KEY` is used). 
        - set_as_default (bool optional): Set the created Datastore as the default datastore for the Workspace.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        workspace: Workspace,
        container_name: str = None,
        datastore_name: str = None,
        create_container_if_not_exists: bool = False,
        overwrite_existing_datastore: bool = False,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        set_as_default: bool = False,
        **kwargs
    ) -> None:
        self.workspace = workspace
        self.container_name = container_name
        self.datastore_name = datastore_name
        self.create_container_if_not_exists = create_container_if_not_exists
        self.overwrite_existing_datastore = overwrite_existing_datastore
        self.azure_credentials_secret = azure_credentials_secret
        self.set_as_default = set_as_default

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "container_name",
        "datastore_name",
        "create_container_if_not_exists",
        "overwrite_existing_datastore",
        "azure_credentials_secret",
        "set_as_default",
    )
    def run(
        self,
        container_name: str = None,
        datastore_name: str = None,
        create_container_if_not_exists: bool = False,
        overwrite_existing_datastore: bool = False,
        azure_credentials_secret: str = "AZ_CREDENTIALS",
        set_as_default: bool = False,
    ) -> AzureBlobDatastore:
        """
        Task run method.
        
        Args:
            - container_name (str, optional): The name of the container.
            - datastore_name (str, optional): The name of the datastore. If not defined, the container name will be used.
            - create_container_if_not_exists (bool, optional): Create a container, if one does not exist with the given name. 
            - overwrite_existing_datastore (bool, optional): Overwrite an existing datastore. If the datastore does not exist, it will be created.
            - azure_credentials_secret (str, optinonal): The name of the Prefect Secret that stores your Azure credentials; this Secret must be a JSON string with two keys: `ACCOUNT_NAME` and either `ACCOUNT_KEY` or `SAS_TOKEN` (if both are defined then`ACCOUNT_KEY` is used)
            - set_as_default (bool optional): Set the created Datastore as the default datastore for the Workspace.

        Return:
            - (azureml.data.azure_storage_datastore.AzureBlobDatastore): The registered Datastore.
        
        """
        if container_name is None:
            raise ValueError("A container name must be provided.")

        if datastore_name is None:
            datastore_name = container_name

        # get Azure credentials
        azure_credentials = Secret(azure_credentials_secret).get()
        az_account_name = azure_credentials["ACCOUNT_NAME"]
        az_account_key = azure_credentials.get("ACCOUNT_KEY")
        az_sas_token = azure_credentials.get("SAS_TOKEN")

        datastore = azureml.core.datastore.Datastore.register_azure_blob_container(
            workspace=self.workspace,
            datastore_name=datastore_name,
            container_name=container_name,
            account_name=az_account_name,
            account_key=az_account_key,
            sas_token=az_sas_token,
            overwrite=overwrite_existing_datastore,
            create_if_not_exists=create_container_if_not_exists,
        )

        if set_as_default:
            datastore.set_as_default()

        return datastore


class DatastoreList(Task):
    """
    Task for listing the Datastores in a Workspace.

    Args:
        - workspace (azureml.core.workspace.Workspace): The Workspace which Datastores are to be listed. 
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(self, workspace: Workspace, **kwargs) -> None:
        self.workspace = workspace

        super().__init__(**kwargs)

    def run(self) -> Dict[str, azureml.core.datastore.Datastore]:
        """
        Task run method.

        Returns:
            -  (Dict[str, Datastore]): a dictionary with the datastore names as keys and Datastore objects as items. 
        """

        return self.workspace.datastores


class DatastoreGet(Task):
    """
    Task for getting a Datastore registered to a given Workspace.

    Args:
        - workspace (azureml.core.workspace.Workspace): The Workspace which Datastore is retrieved. 
        - datastore_name (str, optional): The name of the Datastore. If `None`, then the default Datastore of the Workspace is returned. 
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self, workspace: Workspace, datastore_name: str = None, **kwargs
    ) -> None:
        self.workspace = workspace
        self.datastore_name = datastore_name

        super().__init__(**kwargs)

    @defaults_from_attrs("datastore_name")
    def run(self, datastore_name: str = None) -> azureml.core.datastore.Datastore:
        """
        Task run method. 

        Args:
            - datastore_name (str, optional): The name of the Datastore. If `None`, then the default Datastore of the Workspace is returned. 
    
        Returns:
            - (azureml.core.datastore.Datastore): The Datastore. 

        """
        if datastore_name is None:
            return azureml.core.datastore.Datastore.get_default(self.workspace)
        return azureml.core.datastore.Datastore.get(
            self.workspace, datastore_name=datastore_name
        )


class DatastoreUpload(Task):
    """
    Task for uploading local files to a Datastore.

    Args:
        - datastore (azureml.data.azure_storage_datastore.AbstractAzureStorageDatastore, optional): The datastore to upload the files to. 
        - relative_root (str, optional): The root from which is used to determine the path of the files in the blob. For example, if we upload /path/to/file.txt, and we define base path to be /path, when file.txt is uploaded to the blob storage, it will have the path of /to/file.txt.
        - path (Union[str, List[str]], optional): The path to a single file, single directory, or a list of path to files to eb uploaded. 
        - target_path (str, optional): The location in the blob container to upload to. If None, then upload to root.
        - overwrite (bool, optional): Overwrite existing file(s).
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor

    """

    def __init__(
        self,
        datastore: AbstractAzureStorageDatastore = None,
        relative_root: str = None,
        path: Union[str, List[str]] = None,
        target_path: str = None,
        overwrite: bool = False,
        **kwargs
    ) -> None:
        self.datastore = datastore
        self.path = path
        self.relative_root = relative_root
        self.target_path = target_path
        self.overwrite = overwrite

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "datastore", "path", "relative_root", "target_path", "overwrite"
    )
    def run(
        self,
        datastore: azureml.core.datastore.Datastore = None,
        path: Union[str, List[str]] = None,
        relative_root: str = None,
        target_path: str = None,
        overwrite: bool = False,
    ) -> DataReference:
        """
        Task run method.
        
        Args:
            - datastore (azureml.data.azure_storage_datastore.AbstractAzureStorageDatastore, optional): The datastore to upload the files to. 
            - relative_root (str, optional): The root from which is used to determine the path of the files in the blob. For example, if we upload `/path/to/file.txt`, and we define base path to be `/path`, when `file.txt` is uploaded to the blob storage, it will have the path of `/to/file.txt`.
            - path (Union[str, List[str]], optional): The path to a single file, single directory, or a list of path to files to eb uploaded. 
            - target_path (str, optional): The location in the blob container to upload to. If None, then upload to root.
            - overwrite (bool, optional): Overwrite existing file(s).

        Returns:
            - (azureml.data.data_reference.DataReference): The DataReference instance for the target path uploaded
        """

        if datastore is None:
            raise ValueError("A datastore must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if isinstance(path, str) and os.path.isdir(path):
            data_reference = datastore.upload(
                src_dir=path,
                target_path=target_path,
                overwrite=overwrite,
                show_progress=False,
            )
            return data_reference

        if isinstance(path, str):
            path = [path]

        data_reference = datastore.upload_files(
            files=path,
            relative_root=relative_root,
            target_path=target_path,
            overwrite=overwrite,
            show_progress=False,
        )

        return data_reference



"""
Utility functions for interacting with Google Cloud.
"""
import prefect

from google.oauth2.service_account import Credentials


def get_google_client(submodule, credentials: dict = None, project: str = None):
    """
    Utility function for loading Google Client objects from a given set of credentials.

    Args:
        - submodule: a Python submodule with a Client attribute
        - credentials (dict, optional): a dictionary of Google credentials used to initialize the Client; if
            not provided, will attempt to load the Client using ambient environment settings
        - project (str, optional): the Google project to point the Client to; if not provided, Client defaults
            will be used

    Returns:
        - Client: an initialized and authenticated Google Client
    """
    Client = getattr(submodule, "Client")
    credentials = credentials or prefect.context.get("secrets", {}).get(
        "GCP_CREDENTIALS"
    )
    if credentials is not None:
        credentials = Credentials.from_service_account_info(credentials)
        project = project or credentials.project_id
        client = Client(project=project, credentials=credentials)
    else:
        client = Client(project=project)
    return client


def get_storage_client(credentials: dict = None, project: str = None):
    """
    Utility function for instantiating a Google Storage Client from a given set of credentials.

    Args:
        - credentials (dict, optional): a dictionary of Google credentials used to initialize the Client; if
            not provided, will attempt to load the Client using ambient environment settings
        - project (str, optional): the Google project to point the Client to; if not provided, Client defaults
            will be used

    Returns:
        - Client: an initialized and authenticated Google Client
    """
    from google.cloud import storage

    return get_google_client(storage, credentials=credentials, project=project)


def get_bigquery_client(credentials: dict = None, project: str = None):
    """
    Utility function for instantiating a Google BigQuery Client from a given set of credentials.

    Args:
        - credentials (dict, optional): a dictionary of Google credentials used to initialize the Client; if
            not provided, will attempt to load the Client using ambient environment settings
        - project (str, optional): the Google project to point the Client to; if not provided, Client defaults
            will be used

    Returns:
        - Client: an initialized and authenticated Google Client
    """
    from google.cloud import bigquery

    return get_google_client(bigquery, credentials=credentials, project=project)



from typing import TYPE_CHECKING, Any

from marshmallow import fields, post_load

from prefect.engine import result, state
from prefect.serialization.result import StateResultSchema
from prefect.utilities.serialization import (
    JSONCompatible,
    Nested,
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
)

if TYPE_CHECKING:
    pass


def get_safe(obj: state.State, context: dict) -> Any:
    """
    Helper function for ensuring only safe values are serialized.
    Note that it is up to the user to actively store a Result's value in a
    safe way prior to serialization (if they want the result to be avaiable post-serialization).
    """
    if context.get("attr") == "_result":
        if getattr(obj._result, "location", None) is not None:
            return obj._result
        return obj._result.safe_value  # type: ignore
    value = context.get("value", result.NoResult)
    return value


class BaseStateSchema(ObjectSchema):
    class Meta:
        object_class = state.State

    context = fields.Dict(key=fields.Str(), values=JSONCompatible(), allow_none=True)
    message = fields.String(allow_none=True)
    _result = Nested(StateResultSchema, allow_none=False, value_selection_fn=get_safe)
    cached_inputs = fields.Dict(
        key=fields.Str(),
        values=Nested(StateResultSchema, value_selection_fn=get_safe),
        allow_none=True,
    )

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> state.State:
        result_obj = data.pop("_result", result.NoResult)
        data["result"] = result_obj
        base_obj = super().create_object(data)
        return base_obj


class PendingSchema(BaseStateSchema):
    class Meta:
        object_class = state.Pending


class MetaStateSchema(BaseStateSchema):
    state = fields.Nested("prefect.serialization.state.StateSchema", allow_none=True)


class ClientFailedSchema(MetaStateSchema):
    class Meta:
        object_class = state.ClientFailed


class SubmittedSchema(MetaStateSchema):
    class Meta:
        object_class = state.Submitted


class QueuedSchema(MetaStateSchema):
    class Meta:
        object_class = state.Queued

    start_time = fields.DateTime(allow_none=True)


class ScheduledSchema(PendingSchema):
    class Meta:
        object_class = state.Scheduled

    start_time = fields.DateTime(allow_none=True)


class ResumeSchema(ScheduledSchema):
    class Meta:
        object_class = state.Resume


class RetryingSchema(ScheduledSchema):
    class Meta:
        object_class = state.Retrying

    run_count = fields.Int(allow_none=True)


class RunningSchema(BaseStateSchema):
    class Meta:
        object_class = state.Running


class FinishedSchema(BaseStateSchema):
    class Meta:
        object_class = state.Finished


class LoopedSchema(BaseStateSchema):
    class Meta:
        object_class = state.Looped

    loop_count = fields.Int(allow_none=False)


class SuccessSchema(FinishedSchema):
    class Meta:
        object_class = state.Success


class CachedSchema(SuccessSchema):
    class Meta:
        object_class = state.Cached

    cached_parameters = JSONCompatible(allow_none=True)
    cached_result_expiration = fields.DateTime(allow_none=True)
    hashed_inputs = fields.Dict(key=fields.Str(), values=fields.Str(), allow_none=True)


class MappedSchema(SuccessSchema):
    class Meta:
        exclude = ["_result", "map_states"]
        object_class = state.Mapped

    # though this field is excluded from serialization, it must be present in the schema
    map_states = fields.Nested("prefect.serialization.state.StateSchema", many=True)
    n_map_states = fields.Integer()

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> state.Mapped:
        n_map_states = data.pop("n_map_states", 0)
        data["map_states"] = [None for _ in range(n_map_states)]
        return super().create_object(data)


class FailedSchema(FinishedSchema):
    class Meta:
        object_class = state.Failed


class CancelledSchema(FinishedSchema):
    class Meta:
        object_class = state.Cancelled


class TimedOutSchema(FinishedSchema):
    class Meta:
        object_class = state.TimedOut


class TriggerFailedSchema(FailedSchema):
    class Meta:
        object_class = state.TriggerFailed


class ValidationFailedSchema(FailedSchema):
    class Meta:
        object_class = state.ValidationFailed


class SkippedSchema(SuccessSchema):
    class Meta:
        object_class = state.Skipped
        exclude_fields = ["cached"]


class PausedSchema(ScheduledSchema):
    class Meta:
        object_class = state.Paused


class StateSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "Cancelled": CancelledSchema,
        "Cached": CachedSchema,
        "ClientFailed": ClientFailedSchema,
        "Failed": FailedSchema,
        "Finished": FinishedSchema,
        "Looped": LoopedSchema,
        "Mapped": MappedSchema,
        "Paused": PausedSchema,
        "Pending": PendingSchema,
        "Queued": QueuedSchema,
        "Resume": ResumeSchema,
        "Retrying": RetryingSchema,
        "Running": RunningSchema,
        "Scheduled": ScheduledSchema,
        "Skipped": SkippedSchema,
        "Submitted": SubmittedSchema,
        "Success": SuccessSchema,
        "TimedOut": TimedOutSchema,
        "TriggerFailed": TriggerFailedSchema,
        "ValidationFailed": ValidationFailedSchema,
    }



from typing import Any

from marshmallow import fields, post_load

from prefect.environments import (
    DaskKubernetesEnvironment,
    Environment,
    FargateTaskEnvironment,
    KubernetesJobEnvironment,
    LocalEnvironment,
    RemoteEnvironment,
    RemoteDaskEnvironment,
)
from prefect.utilities.serialization import (
    ObjectSchema,
    OneOfSchema,
    to_qualified_name,
    JSONCompatible,
)


class BaseEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = Environment

    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class LocalEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = LocalEnvironment

    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class DaskKubernetesEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = DaskKubernetesEnvironment

    docker_secret = fields.String(allow_none=True)
    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)
    private_registry = fields.Boolean(allow_none=False)
    min_workers = fields.Int()
    max_workers = fields.Int()


class FargateTaskEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = FargateTaskEnvironment

    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class KubernetesJobEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = KubernetesJobEnvironment

    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class RemoteEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = RemoteEnvironment

    executor = fields.String(allow_none=True)
    executor_kwargs = fields.Dict(allow_none=True)
    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class RemoteDaskEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = RemoteDaskEnvironment

    address = fields.String()
    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)


class CustomEnvironmentSchema(ObjectSchema):
    class Meta:
        object_class = lambda: Environment
        exclude_fields = ["type"]

    labels = fields.List(fields.String())
    metadata = JSONCompatible(allow_none=True)

    type = fields.Function(
        lambda environment: to_qualified_name(type(environment)), lambda x: x
    )

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Environment:
        """
        Because we cannot deserialize a custom class, we return an empty
        Base Environment with the appropriate labels.
        """
        return Environment(labels=data.get("labels"), metadata=data.get("metadata"))


class EnvironmentSchema(OneOfSchema):
    """
    Field that chooses between several nested schemas
    """

    # map class name to schema
    type_schemas = {
        "DaskKubernetesEnvironment": DaskKubernetesEnvironmentSchema,
        "Environment": BaseEnvironmentSchema,
        "FargateTaskEnvironment": FargateTaskEnvironmentSchema,
        "LocalEnvironment": LocalEnvironmentSchema,
        "KubernetesJobEnvironment": KubernetesJobEnvironmentSchema,
        "RemoteEnvironment": RemoteEnvironmentSchema,
        "RemoteDaskEnvironment": RemoteDaskEnvironmentSchema,
        "CustomEnvironment": CustomEnvironmentSchema,
    }

    def get_obj_type(self, obj: Any) -> str:
        name = obj.__class__.__name__
        if name in self.type_schemas:
            return name
        else:
            return "CustomEnvironment"



"""
Utility functions for interacting with and configuring logging.  The main entrypoint for retrieving loggers for
customization is the `get_logger` utility.

Note that Prefect Tasks come equipped with their own loggers.  These can be accessed via:
    - `self.logger` if implementing a Task class
    - `prefect.context.get("logger")` if using the `task` decorator

When running locally, log levels and message formatting are set via your Prefect configuration file.
"""
import atexit
import json
import logging
import sys
import threading
import time
from queue import Empty, Queue
from typing import Any

import pendulum

import prefect
from prefect.utilities.context import context

_original_log_record_factory = logging.getLogRecordFactory()

PREFECT_LOG_RECORD_ATTRIBUTES = (
    "flow_name",
    "flow_run_id",
    "task_name",
    "task_slug",
    "task_run_id",
)


class CloudHandler(logging.StreamHandler):
    def __init__(self) -> None:
        super().__init__(sys.stdout)
        self.client = None
        self.logger = logging.getLogger("CloudHandler")
        handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            context.config.logging.format, context.config.logging.datefmt
        )
        formatter.converter = time.gmtime  # type: ignore
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(context.config.logging.level)

    @property
    def queue(self) -> Queue:
        if not hasattr(self, "_queue"):
            self._queue = Queue()  # type: Queue
            self._flush = False
            self.start()
        return self._queue

    def flush(self) -> None:
        self._flush = True
        if self.client is not None:
            self.batch_upload()
            self._thread.join()

    def batch_upload(self) -> None:
        logs = []
        try:
            while True:
                log = self.queue.get(False)
                logs.append(log)
        except Empty:
            pass

        if logs:
            try:
                assert self.client is not None
                self.client.write_run_logs(logs)
            except Exception as exc:
                message = "Failed to write log with error: {}".format(str(exc))
                self.logger.critical(message)

                # Attempt to write batch error log otherwise log invalid cloud communication
                try:
                    assert self.client is not None
                    self.client.write_run_logs([self._make_error_log(message)])
                except Exception as exc:
                    self.logger.critical("Unable to write logs to Prefect Cloud")

    def _monitor(self) -> None:
        while not self._flush:
            self.batch_upload()
            time.sleep(self.heartbeat)

    def __del__(self) -> None:
        if hasattr(self, "_thread"):
            self.flush()
            atexit.unregister(self.flush)

    def start(self) -> None:
        if not hasattr(self, "_thread"):
            self.heartbeat = context.config.cloud.logging_heartbeat
            self._thread = t = threading.Thread(
                target=self._monitor, name="PrefectCloudLoggingThread"
            )
            t.daemon = True
            t.start()
            atexit.register(self.flush)

    def put(self, log: dict) -> None:
        try:
            json.dumps(log)  # make sure the payload is serializable
            self.queue.put(log)
        except TypeError as exc:
            message = "Failed to write log with error: {}".format(str(exc))
            self.logger.critical(message)

            self.queue.put(self._make_error_log(message))

    def emit(self, record) -> None:  # type: ignore
        # if we shouldn't log to cloud, don't emit
        if not prefect.context.config.logging.log_to_cloud:
            return

        try:
            from prefect.client import Client

            if self.client is None:
                self.client = Client()  # type: ignore

            assert isinstance(self.client, Client)  # mypy assert

            record_dict = record.__dict__.copy()
            ## remove potentially non-json serializable formatting args
            record_dict.pop("args", None)

            log = dict()
            log["flow_run_id"] = prefect.context.get("flow_run_id", None)
            log["task_run_id"] = prefect.context.get("task_run_id", None)
            log["timestamp"] = pendulum.from_timestamp(
                record_dict.pop("created", time.time())
            ).isoformat()
            log["name"] = record_dict.pop("name", None)
            log["message"] = record_dict.pop("message", None)
            log["level"] = record_dict.pop("levelname", None)

            if record_dict.get("exc_text") is not None:
                log["message"] += "\n" + record_dict.pop("exc_text", "")
                record_dict.pop("exc_info", None)

            log["info"] = record_dict
            self.put(log)
        except Exception as exc:
            message = "Failed to write log with error: {}".format(str(exc))
            self.logger.critical(message)

            self.put(self._make_error_log(message))

    def _make_error_log(self, message: str) -> dict:
        log = dict()
        log["flow_run_id"] = prefect.context.get("flow_run_id", None)
        log["timestamp"] = pendulum.from_timestamp(time.time()).isoformat()
        log["name"] = self.logger.name
        log["message"] = message
        log["level"] = "CRITICAL"
        log["info"] = {}

        return log


def _log_record_context_injector(*args: Any, **kwargs: Any) -> logging.LogRecord:
    """
    A custom logger LogRecord Factory that injects selected context parameters into newly created logs.

    Args:
        - *args: arguments to pass to the original LogRecord Factory
        - **kwargs: keyword arguments to pass to the original LogRecord Factory

    Returns:
        - logging.LogRecord: the newly created LogRecord
    """
    record = _original_log_record_factory(*args, **kwargs)

    additional_attrs = context.config.logging.get("log_attributes", [])

    for attr in PREFECT_LOG_RECORD_ATTRIBUTES + tuple(additional_attrs):
        value = prefect.context.get(attr, None)
        if value or attr in additional_attrs:
            setattr(record, attr, value)

    return record


def _create_logger(name: str) -> logging.Logger:
    """
    Creates a logger with a `StreamHandler` that has level and formatting
    set from `prefect.config`.

    Args:
        - name (str): Name to use for logger.

    Returns:
        - logging.Logger: a configured logging object
    """
    logging.setLogRecordFactory(_log_record_context_injector)

    logger = logging.getLogger(name)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(
        context.config.logging.format, context.config.logging.datefmt
    )
    formatter.converter = time.gmtime  # type: ignore
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(context.config.logging.level)

    cloud_handler = CloudHandler()
    cloud_handler.setLevel("DEBUG")
    logger.addHandler(cloud_handler)
    return logger


def configure_logging(testing: bool = False) -> logging.Logger:
    """
    Creates a "prefect" root logger with a `StreamHandler` that has level and formatting
    set from `prefect.config`.

    Args:
        - testing (bool, optional): a boolean specifying whether this configuration
            is for testing purposes only; this helps us isolate any global state during testing
            by configuring a "prefect-test-logger" instead of the standard "prefect" logger

    Returns:
        - logging.Logger: a configured logging object
    """
    name = "prefect-test-logger" if testing else "prefect"
    return _create_logger(name)


prefect_logger = configure_logging()


def configure_extra_loggers() -> None:
    """
    Creates a "Prefect" configured logger for all strings in extra_loggers config list.
    The logging.extra_loggers config defaults to an empty list.
    """
    loggers = context.config.logging.get("extra_loggers", [])
    for l in loggers:
        _create_logger(l)


configure_extra_loggers()


def create_diagnostic_logger(name: str) -> logging.Logger:
    """
    Create a logger that does not use the `CloudHandler` but preserves all other
    Prefect logging configuration.  For diagnostic / debugging / internal use only.
    """
    logger = _create_logger(name)
    logger.handlers = [h for h in logger.handlers if not isinstance(h, CloudHandler)]
    return logger


def get_logger(name: str = None) -> logging.Logger:
    """
    Returns a "prefect" logger.

    Args:
        - name (str): if `None`, the root Prefect logger is returned. If provided, a child
            logger of the name `"prefect.{name}"` is returned. The child logger inherits
            the root logger's settings.

    Returns:
        - logging.Logger: a configured logging object with the appropriate name
    """

    if name is None:
        return prefect_logger
    else:
        return prefect_logger.getChild(name)


class RedirectToLog:
    """
    Custom redirect of stdout messages to logs

    Args:
        - logger (logging.Logger, optional): an optional logger to redirect stdout. If
            not provided a logger names `stdout` will be created.
    """

    def __init__(self, logger: logging.Logger = None) -> None:
        self.stdout_logger = logger or get_logger("stdout")

    def write(self, s: str) -> None:
        """
        Write message from stdout to a prefect logger.
        Note: blank newlines will not be logged.

        Args:
            s (str): the message from stdout to be logged
        """
        if s.strip():
            self.stdout_logger.info(s)

    def flush(self) -> None:
        """
        Implemented flush operation for logger handler
        """
        for handler in self.stdout_logger.handlers:
            handler.flush()



import base64
import datetime
import inspect
import json
import sys
import types
from typing import Any, Callable, List

import marshmallow_oneofschema
import pendulum
from marshmallow import (
    EXCLUDE,
    Schema,
    SchemaOpts,
    ValidationError,
    fields,
    missing,
    post_dump,
    post_load,
    pre_load,
)

import prefect


def to_qualified_name(obj: Any) -> str:
    """
    Given an object, returns its fully-qualified name, meaning a string that represents its
    Python import path

    Args:
        - obj (Any): an importable Python object

    Returns:
        - str: the qualified name
    """
    return obj.__module__ + "." + obj.__qualname__


def from_qualified_name(obj_str: str) -> Any:
    """
    Retrives an object from a fully qualified string path. The object must be
    imported in advance.

    Args:
        - obj_str (str): the qualified path of the object

    Returns:
        - Any: the object retrieved from the qualified path

    Raises:
        - ValueError: if the object could not be loaded from the supplied path. Note that
            this function will not import objects; they must be imported in advance.
    """

    path_components = obj_str.split(".")
    try:
        for i in range(len(path_components), 0, -1):
            module_path = ".".join(path_components[:i])
            if module_path in sys.modules:
                obj = sys.modules[module_path]
                for p in path_components[i:]:
                    obj = getattr(obj, p)
                return obj
    except Exception:
        pass  # exceptions are raised by the catch-all at the end
    raise ValueError(
        "Couldn't load \"{}\"; maybe it hasn't been imported yet?".format(obj_str)
    )


class ObjectSchemaOptions(SchemaOpts):
    def __init__(self, meta: Any, **kwargs: Any) -> None:
        super().__init__(meta, **kwargs)
        self.object_class = getattr(meta, "object_class", None)
        self.exclude_fields = getattr(meta, "exclude_fields", None) or []
        self.unknown = getattr(meta, "unknown", EXCLUDE)


class ObjectSchema(Schema):
    """
    This Marshmallow Schema automatically instantiates an object whose type is indicated by the
    `object_class` attribute of the class `Meta`. All deserialized fields are passed to the
    constructor's `__init__()` unless the name of the field appears in `Meta.exclude_fields`.
    """

    OPTIONS_CLASS = ObjectSchemaOptions

    class Meta:
        object_class = None  # type: type
        exclude_fields = []  # type: List[str]
        unknown = EXCLUDE

    @pre_load
    def _remove_version(self, data: dict, **kwargs: Any) -> dict:
        """
        Removes a __version__ field from the data, if present.

        Args:
            - data (dict): the serialized data

        Returns:
            - dict: the data dict, without its __version__ field
        """
        # don't mutate data
        data = data.copy()
        data.pop("__version__", None)
        return data

    @post_dump
    def _add_version(self, data: dict, **kwargs: Any) -> dict:
        """
        Adds a __version__ field to the data, if not already provided.

        Args:
            - data (dict): the serialized data

        Returns:
            - dict: the data dict, with an additional __version__ field
        """
        # don't mutate data
        data = data.copy()
        data.setdefault("__version__", prefect.__version__)
        return data

    def load(self, data: dict, create_object: bool = True, **kwargs: Any) -> Any:
        """
        Loads an object by first retrieving the appropate schema version (based on the data's
        __version__ key).

        Args:
            - data (dict): the serialized data
            - create_object (bool): if True, an instantiated object will be returned. Otherwise,
                the deserialized data dict will be returned.
            - **kwargs (Any): additional keyword arguments for the load() method

        Returns:
            - Any: the deserialized object or data
        """
        self.context.setdefault("create_object", create_object)
        return super().load(data, **kwargs)

    @post_load
    def create_object(self, data: dict, **kwargs: Any) -> Any:
        """
        By default, returns an instantiated object using the ObjectSchema's `object_class`.
        Otherwise, returns a data dict.

        Args:
            - data (dict): the deserialized data

        Returns:
            - Any: an instantiated object, if `create_object` is found in context; otherwise, the data dict
        """
        if self.context.get("create_object", True):
            object_class = self.opts.object_class
            if object_class is not None:
                if isinstance(object_class, types.FunctionType):
                    object_class = object_class()
                init_data = {
                    k: v for k, v in data.items() if k not in self.opts.exclude_fields
                }
                return object_class(**init_data)
        return data


class JSONCompatible(fields.Field):
    """
    Field that ensures its values are JSON-compatible during serialization and deserialization

    Args:
        - *args (Any): the arguments accepted by `marshmallow.Field`
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.validators.insert(0, self._validate_json)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        self._validate_json(value)
        return super()._serialize(value, attr, obj, **kwargs)

    def _validate_json(self, value: Any) -> None:
        try:
            json.dumps(value)
        except TypeError:
            raise ValidationError("Value is not JSON-compatible")


class Nested(fields.Nested):
    """
    An extension of the Marshmallow Nested field that allows the value to be selected
    via a value_selection_fn.

    Note that because the value_selection_fn is always called, users must return
    `marshmallow.missing` if they don't want this field included in the resulting serialized
    object.

    Args:
        - nested (type): the nested schema class
        - value_selection_fn (Callable): a function that is called whenever the object is serialized,
            to retrieve the object (if not available as a simple attribute of the parent schema)
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`
    """

    def __init__(self, nested: type, value_selection_fn: Callable, **kwargs: Any):
        super().__init__(nested=nested, **kwargs)
        self.value_selection_fn = value_selection_fn

        if value_selection_fn is not None:
            self._CHECK_ATTRIBUTE = False

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if self.value_selection_fn is not None:
            self.context.update(value=value, attr=attr)
            value = self.value_selection_fn(obj, self.context)
        if value is missing:
            return value
        return super()._serialize(value, attr, obj, **kwargs)


class OneOfSchema(marshmallow_oneofschema.OneOfSchema):
    """
    A subclass of marshmallow_oneofschema.OneOfSchema that excludes unknown fields
    """

    class Meta:
        unknown = EXCLUDE


class Bytes(fields.Field):
    """
    A Marshmallow Field that serializes bytes to a base64-encoded string, and deserializes
    a base64-encoded string to bytes.

    Args:
        - *args (Any): the arguments accepted by `marshmallow.Field`
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if value is not None:
            return base64.b64encode(value).decode("utf-8")

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        if value is not None:
            return base64.b64decode(value)


class UUID(fields.UUID):
    """
    Replacement for fields.UUID that performs validation but returns string objects,
    not UUIDs

    Args:
        - *args (Any): the arguments accepted by `marshmallow.Field`
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        return super()._serialize(value, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        return str(super()._deserialize(value, attr, data, **kwargs))


class DateTimeTZ(fields.DateTime):
    """
    Replacement for fields.DateTime that stores the time zone (not the time zone offset like
    fields.LocalDateTime)

    Args:
        - *args (Any): the arguments accepted by `marshmallow.Field`
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if value is not None:
            dt = pendulum.instance(value)
            return dict(dt=dt.naive().to_iso8601_string(), tz=dt.tzinfo.name)

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        if value is not None:
            return pendulum.parse(value["dt"], tz=value["tz"])


class FunctionReference(fields.Field):
    """
    Field that stores a reference to a function as a string and reloads it when
    deserialized.

    Args:
        - valid_functions (List[Callable]): a list of functions that will be serialized as string
            references
        - reject_invalid (bool): if True, functions not in `valid_functions` will be rejected. If False,
            any value will be allowed, but only functions in `valid_functions` will be deserialized.
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`

    """

    def __init__(
        self, valid_functions: list, reject_invalid: bool = True, **kwargs: Any
    ):
        self.valid_functions = {to_qualified_name(f): f for f in valid_functions}
        self.reject_invalid = reject_invalid
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if value is None and self.allow_none:
            return None
        elif value not in self.valid_functions.values() and self.reject_invalid:
            raise ValidationError("Invalid function reference: {}".format(value))
        return to_qualified_name(value)

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        if value is None and self.allow_none:
            return None
        elif value not in self.valid_functions and self.reject_invalid:
            raise ValidationError("Invalid function reference: {}".format(value))
        return self.valid_functions.get(value, value)


class StatefulFunctionReference(fields.Field):
    """

    Field that stores a reference to a function as a string and reloads it when
    deserialized.

    In addition, if the function was generated by another function, like a factory, then this
    field will store a reference to the factory function and call it with any `kwargs` required
    by the inner function.
    For example:

    ```python
    def outer(x, y):
        def inner(z):
            return x + y + z
        return inner

    f = outer(1, 2)
    ```

    Here, `f` is a "stateful" function because it depends on the values `x` and `y`, but has
    no explicit attribute holding them. It was generated by a closure that requires them.

    We would serialize this function as `{"fn": "outer", "kwargs": {"x": 1, "y": 2}}`

    Note that `nonlocals` MUST be JSON-compatible, with the exception of datetimes.

    Args:
        - valid_functions (List[Callable]): an allow list of valid functions
        - reject_invalid (bool): if True, functions not in `valid_functions` will be rejected. If False,
            any value will be allowed, but only functions in `valid_functions` will be deserialized.
        - **kwargs (Any): the keyword arguments accepted by `marshmallow.Field`

    """

    def __init__(
        self, valid_functions: list, reject_invalid: bool = True, **kwargs: Any
    ):
        self.valid_functions = {to_qualified_name(f): f for f in valid_functions}
        self.reject_invalid = reject_invalid
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, **kwargs):  # type: ignore
        if value is None and self.allow_none:
            return None

        try:
            qual_name = to_qualified_name(value)
        except:
            raise ValidationError(
                f"Invalid function reference, function required, got {value}"
            )

        # sort matches such that the longest / most specific match comes first
        valid_bases = sorted(
            [fn for fn in self.valid_functions if qual_name.startswith(fn)],
            key=lambda k: len(k),
            reverse=True,
        )
        if not valid_bases and self.reject_invalid:
            raise ValidationError("Invalid function reference: {}".format(value))
        else:
            base_name = next(filter(None, valid_bases)) if valid_bases else qual_name

        nonlocals = dict(inspect.getclosurevars(value).nonlocals)

        for k, v in list(nonlocals.items()):
            # convert dates to strings
            if isinstance(v, datetime.datetime):
                nonlocals[k] = "//datetime:" + v.isoformat()
            # convert time to strings
            elif isinstance(v, datetime.time):
                nonlocals[k] = "//time:" + v.isoformat()
            # convert timedelta to seconds
            elif isinstance(v, datetime.timedelta):
                nonlocals[k] = "//timedelta:" + str(v.total_seconds())

        return {"fn": base_name, "kwargs": nonlocals}

    def _deserialize(self, value, attr, data, **kwargs):  # type: ignore
        """
        Attempts to return the original function that was serialized, with _no state_.

        If not found, returns `None`.
        """
        if value is None and self.allow_none:
            return None

        # retrieve the function
        base_name = value["fn"]
        if base_name not in self.valid_functions and self.reject_invalid:
            raise ValidationError("Invalid function reference: {}".format(value))
        fn = self.valid_functions.get(base_name)

        if not fn:
            return None

        # call function on state
        kwargs = value.get("kwargs", {}).copy()

        # if there are no kwargs, then this function isn't stateful
        if not kwargs:
            return fn

        # if there ARE kwargs, then this function is actually the parent function and
        # it needs to be called
        for k, v in list(kwargs.items()):
            # parse datetimes
            if isinstance(v, str) and v.startswith("//datetime:"):
                kwargs[k] = pendulum.parse(v.split("//datetime:", 1)[1])
            # parse times
            if isinstance(v, str) and v.startswith("//time:"):
                kwargs[k] = pendulum.parse(v.split("//time:", 1)[1]).time()
            # parse timedeltas
            if isinstance(v, str) and v.startswith("//timedelta:"):
                kwargs[k] = datetime.timedelta(
                    seconds=float(v.split("//timedelta:", 1)[1])
                )
        return fn(**kwargs)



from typing import Dict, List, Union

import azureml.core.dataset
from azureml.core.datastore import Datastore
from azureml.data import DataType, TabularDataset
from azureml.data.dataset_type_definitions import PromoteHeadersBehavior

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs


class DatasetCreateFromDelimitedFiles(Task):
    """
    Task for creating a TabularDataset from delimited files for use in a Azure Machine Learning service Workspace. The files should exist in a Datastore.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - include_path (bool, optional): Boolean to keep path information as column in the dataset.
        - infer_column_types (bool, optional): Boolean to infer column data types.
        - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set column data type, where key is column name and value is a `azureml.data.DataType`.
        - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
        - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
        - separator (str, optional): The separator used to split columns.
        - header (azureml.data.dataset_type_definitions.PromoteHeadersBehavior, optional): Controls how column headers are promoted when reading from files. Defaults to assume that all files have the same header.
        - partition_format (str, optional): Specify the partition format of path. Defaults to None. The partition information of each path will be extracted into columns based on the specified format. Format part `{column_name}` creates string column, and `{column_name:yyyy/MM/dd/HH/mm/ss}` creates datetime column, where `yyyy`, `MM`, `dd`, `HH`, `mm` and `ss` are used to extrat year, month, day, hour, minute and second for the datetime type. The format should start from the position of first partition key until the end of file path. For example, given the path `../Germany/2019/01/01/data.csv` where the partition is by country and time, `partition_format="/{Country}/{PartitionDate:yyyy/MM/dd}/data.csv"` creates string column `Country` with value `Germany` and datetime column `PartitionDate` with value `2019-01-01`.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the Task constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        include_path: bool = False,
        infer_column_types: bool = True,
        set_column_types: Dict[str, DataType] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        separator: str = ",",
        header: PromoteHeadersBehavior = PromoteHeadersBehavior.ALL_FILES_HAVE_SAME_HEADERS,
        partition_format: str = None,
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags
        self.include_path = include_path
        self.infer_column_types = infer_column_types
        self.set_column_types = set_column_types
        self.fine_grain_timestamp = fine_grain_timestamp
        self.coarse_grain_timestamp = coarse_grain_timestamp
        self.separator = separator
        self.header = header
        self.partition_format = partition_format
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "include_path",
        "infer_column_types",
        "set_column_types",
        "fine_grain_timestamp",
        "coarse_grain_timestamp",
        "separator",
        "header",
        "partition_format",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        include_path: bool = False,
        infer_column_types: bool = True,
        set_column_types: Dict[str, DataType] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        separator: str = ",",
        header: PromoteHeadersBehavior = PromoteHeadersBehavior.ALL_FILES_HAVE_SAME_HEADERS,
        partition_format: str = None,
        create_new_version: bool = False,
    ) -> TabularDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - include_path (bool, optional): Boolean to keep path information as column in the dataset.
            - infer_column_types (bool, optional): Boolean to infer column data types.
            - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set column data type, where key is column name and value is a `azureml.data.DataType`.
            - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
            - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
            - separator (str, optional): The separator used to split columns.
            - header (azureml.data.dataset_type_definitions.PromoteHeadersBehavior, optional): Controls how column headers are promoted when reading from files. Defaults to assume that all files have the same header.
            - partition_format (str, optional): Specify the partition format of path.
            - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.

        Returns:
            - azureml.data.TabularDataset: the created TabularDataset

        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset = azureml.core.dataset.Dataset.Tabular.from_delimited_files(
            path=[(datastore, path_item) for path_item in path],
            include_path=include_path,
            infer_column_types=infer_column_types,
            set_column_types=set_column_types,
            separator=separator,
            header=header,
            partition_format=partition_format,
        )

        dataset = dataset.with_timestamp_columns(
            fine_grain_timestamp=fine_grain_timestamp,
            coarse_grain_timestamp=coarse_grain_timestamp,
            validate=True,
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset


class DatasetCreateFromParquetFiles(Task):
    """
    Task for creating a TabularDataset from Parquet files for use in a Azure Machine Learning service Workspace. The files should exist in a Datastore.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - include_path (bool, optional): Boolean to keep path information as column in the dataset.
        - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set column data type, where key is column name and value is a `azureml.data.DataType`.
        - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
        - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
        - partition_format (str, optional): Specify the partition format of path. Defaults to None. The partition information of each path will be extracted into columns based on the specified format. Format part `{column_name}` creates string column, and `{column_name:yyyy/MM/dd/HH/mm/ss}` creates datetime column, where `yyyy`, `MM`, `dd`, `HH`, `mm` and `ss` are used to extrat year, month, day, hour, minute and second for the datetime type. The format should start from the position of first partition key until the end of file path. For example, given the path `../Germany/2019/01/01/data.csv` where the partition is by country and time, `partition_format="/{Country}/{PartitionDate:yyyy/MM/dd}/data.csv"` creates string column `Country` with value `Germany` and datetime column `PartitionDate` with value `2019-01-01`.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        include_path: bool = False,
        set_column_types: Dict[str, DataType] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        partition_format: str = None,
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags
        self.include_path = include_path
        self.set_column_types = set_column_types
        self.fine_grain_timestamp = fine_grain_timestamp
        self.coarse_grain_timestamp = coarse_grain_timestamp
        self.partition_format = partition_format
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "include_path",
        "set_column_types",
        "fine_grain_timestamp",
        "coarse_grain_timestamp",
        "partition_format",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        include_path: bool = False,
        set_column_types: Dict[str, DataType] = None,
        fine_grain_timestamp: str = None,
        coarse_grain_timestamp: str = None,
        partition_format: str = None,
        create_new_version: bool = None,
    ) -> TabularDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - include_path (bool, optional): Boolean to keep path information as column in the dataset.
            - set_column_types (Dict[str, azureml.data.DataType], optional): A dictionary to set column data type, where key is column name and value is a `azureml.data.DataType`.
            - fine_grain_timestamp (str, optional): The name of column as fine grain timestamp.
            - coarse_grain_timestamp (str, optional): The name of column coarse grain timestamp.
            - partition_format (str, optional): Specify the partition format of path.
            - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.

        Returns:
            - azureml.data.TabularDataset: the created TabularDataset.
        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset = azureml.core.dataset.Dataset.Tabular.from_parquet_files(
            path=[(datastore, path_item) for path_item in path],
            include_path=include_path,
            set_column_types=set_column_types,
            partition_format=partition_format,
        )

        dataset = dataset.with_timestamp_columns(
            fine_grain_timestamp=fine_grain_timestamp,
            coarse_grain_timestamp=coarse_grain_timestamp,
            validate=True,
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset


class DatasetCreateFromFiles(Task):
    """
    Task for creating a FileDataset from files for use in a Azure Machine Learning service Workspace. The files should exist in a Datastore.
    Note that all initialization arguments can optionally be provided or overwritten at runtime.

    Args:
        - dataset_name (str, optional): The name of the Dataset in the Workspace
        - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
        - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
        - dataset_description (str, optional): Description of the Dataset.
        - dataset_tags (str, optional): Tags to associate with the Dataset.
        - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        create_new_version: bool = False,
        **kwargs
    ) -> None:
        self.dataset_name = dataset_name
        self.datastore = datastore
        self.path = path
        self.dataset_description = dataset_description
        self.dataset_tags = dataset_tags
        self.create_new_version = create_new_version

        super().__init__(**kwargs)

    @defaults_from_attrs(
        "dataset_name",
        "datastore",
        "path",
        "dataset_description",
        "dataset_tags",
        "create_new_version",
    )
    def run(
        self,
        dataset_name: str = None,
        datastore: Datastore = None,
        path: Union[str, List[str]] = None,
        dataset_description: str = "",
        dataset_tags: Dict[str, str] = dict(),
        create_new_version: bool = False,
    ) -> azureml.data.FileDataset:
        """
        Task run method.

        Args:
            - dataset_name (str, optional): The name of the Dataset in the Workspace
            - datastore (azureml.core.datastore.Datastore, optional): The Datastore which holds the files.
            - path (Union[str, List[str]], optional): The path to the delimited files in the Datastore.
            - dataset_description (str, optional): Description of the Dataset.
            - dataset_tags (str, optional): Tags to associate with the Dataset.
            - create_new_version (bool, optional): Boolean to register the dataset as a new version under the specified name.

        Returns:
            - azureml.data.FileDataset: the created FileDataset
        """
        if dataset_name is None:
            raise ValueError("A dataset_name must be provided.")

        if path is None:
            raise ValueError("A path must be provided.")

        if datastore is None:
            raise ValueError("A path must be provided.")

        if not isinstance(path, list):
            path = [path]

        dataset = azureml.core.dataset.Dataset.File.from_files(
            path=[(datastore, path_item) for path_item in path]
        )

        dataset = dataset.register(
            workspace=datastore.workspace,
            name=dataset_name,
            description=dataset_description,
            tags=dataset_tags,
            create_new_version=create_new_version,
        )

        return dataset



import functools
import logging
import math
import os
import signal
import sys
import threading
import time
from concurrent.futures import Future, ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Generator, Iterable, Set, Optional, cast
from urllib.parse import urlparse

import pendulum
from tornado import web
from tornado.ioloop import IOLoop

from prefect import config
from prefect.client import Client
from prefect.engine.state import Failed, Submitted
from prefect.serialization import state
from prefect.utilities.context import context
from prefect.utilities.exceptions import AuthorizationError
from prefect.utilities.graphql import GraphQLResult, with_args

ascii_name = r"""
 ____            __           _        _                    _
|  _ \ _ __ ___ / _| ___  ___| |_     / \   __ _  ___ _ __ | |_
| |_) | '__/ _ \ |_ / _ \/ __| __|   / _ \ / _` |/ _ \ '_ \| __|
|  __/| | |  __/  _|  __/ (__| |_   / ___ \ (_| |  __/ | | | |_
|_|   |_|  \___|_|  \___|\___|\__| /_/   \_\__, |\___|_| |_|\__|
                                           |___/
"""


@contextmanager
def exit_handler(agent: "Agent") -> Generator:
    exit_event = threading.Event()

    def _exit_handler(*args: Any, **kwargs: Any) -> None:
        agent.logger.info("Keyboard Interrupt received: Agent is shutting down.")
        exit_event.set()

    original = signal.getsignal(signal.SIGINT)
    try:
        signal.signal(signal.SIGINT, _exit_handler)
        yield exit_event
    except SystemExit:
        pass
    finally:
        signal.signal(signal.SIGINT, original)


class HealthHandler(web.RequestHandler):
    """Respond to /api/health"""

    def get(self) -> None:
        # Empty json blob, may add more info later
        self.write({})


class Agent:
    """
    Base class for Agents. Information on using the Prefect agents can be found at
    https://docs.prefect.io/orchestration/agents/overview.html

    This Agent class is a standard point for executing Flows in Prefect Cloud. It is meant
    to have subclasses which inherit functionality from this class. The only piece that
    the subclasses should implement is the `deploy_flows` function, which specifies how to run a Flow on the given platform. It is built in this
    way to keep Prefect Cloud logic standard but allows for platform specific
    customizability.

    In order for this to operate `PREFECT__CLOUD__AGENT__AUTH_TOKEN` must be set as an
    environment variable or in your user configuration file.

    Args:
        - name (str, optional): An optional name to give this agent. Can also be set through
            the environment variable `PREFECT__CLOUD__AGENT__NAME`. Defaults to "agent"
        - labels (List[str], optional): a list of labels, which are arbitrary string identifiers used by Prefect
            Agents when polling for work
        - env_vars (dict, optional): a dictionary of environment variables and values that will be set
            on each flow run that this agent submits for execution
        - max_polls (int, optional): maximum number of times the agent will poll Prefect Cloud for flow runs;
            defaults to infinite
        - agent_address (str, optional): Address to serve internal api at. Currently this is
            just health checks for use by an orchestration layer. Leave blank for no api server (default).
        - no_cloud_logs (bool, optional): Disable logging to a Prefect backend for this agent and all deployed flow runs
    """

    def __init__(
        self,
        name: str = None,
        labels: Iterable[str] = None,
        env_vars: dict = None,
        max_polls: int = None,
        agent_address: str = None,
        no_cloud_logs: bool = False,
    ) -> None:
        self.name = name or config.cloud.agent.get("name", "agent")

        self.labels = labels or list(config.cloud.agent.get("labels", []))
        self.env_vars = env_vars or config.cloud.agent.get("env_vars", dict())
        self.max_polls = max_polls
        self.log_to_cloud = False if no_cloud_logs else True

        self.agent_address = agent_address or config.cloud.agent.get(
            "agent_address", ""
        )
        self._api_server = None  # type: ignore
        self._api_server_loop = None  # type: Optional[IOLoop]
        self._api_server_thread = None  # type: Optional[threading.Thread]

        logger = logging.getLogger(self.name)
        logger.setLevel(config.cloud.agent.get("level"))
        if not any([isinstance(h, logging.StreamHandler) for h in logger.handlers]):
            ch = logging.StreamHandler(sys.stdout)
            formatter = logging.Formatter(context.config.logging.format)
            formatter.converter = time.gmtime  # type: ignore
            ch.setFormatter(formatter)
            logger.addHandler(ch)

        self.logger = logger
        self.submitting_flow_runs = set()  # type: Set[str]

        self.logger.debug("Verbose logs enabled")
        self.logger.debug(f"Environment variables: {[*self.env_vars]}")
        self.logger.debug(f"Max polls: {self.max_polls}")
        self.logger.debug(f"Agent address: {self.agent_address}")
        self.logger.debug(f"Log to Cloud: {self.log_to_cloud}")

        token = config.cloud.agent.get("auth_token")

        self.logger.debug(f"Prefect backend: {config.backend}")

        self.client = Client(api_token=token)

    def _verify_token(self, token: str) -> None:
        """
        Checks whether a token with a `RUNNER` scope was provided
        Args:
            - token (str): The provided agent token to verify
        Raises:
            - AuthorizationError: if token is empty or does not have a RUNNER role
        """
        if not token:
            raise AuthorizationError("No agent API token provided.")

        # Check if RUNNER role
        result = self.client.graphql(query="query { auth_info { api_token_scope } }")
        if (
            not result.data  # type: ignore
            or result.data.auth_info.api_token_scope != "RUNNER"  # type: ignore
        ):
            raise AuthorizationError("Provided token does not have a RUNNER scope.")

    def _register_agent(self) -> str:
        """
        Register this agent with Prefect Cloud and retrieve agent ID

        Returns:
            - The agent ID as a string
        """
        agent_id = self.client.register_agent(
            agent_type=type(self).__name__, name=self.name, labels=self.labels  # type: ignore
        )

        self.logger.debug(f"Agent ID: {agent_id}")

        return agent_id

    def start(self) -> None:
        """
        The main entrypoint to the agent. This function loops and constantly polls for
        new flow runs to deploy
        """
        if config.backend == "cloud":
            self._verify_token(self.client.get_auth_token())
            self.client.attach_headers({"X-PREFECT-AGENT-ID": self._register_agent()})

        try:
            self.setup()

            with exit_handler(self) as exit_event:

                # Loop intervals for query sleep backoff
                loop_intervals = {
                    0: 0.25,
                    1: 0.5,
                    2: 1.0,
                    3: 2.0,
                    4: 4.0,
                    5: 8.0,
                    6: 10.0,
                }

                index = 0
                remaining_polls = math.inf if self.max_polls is None else self.max_polls

                # the max workers default has changed in 3.8. For stable results the
                # default 3.8 behavior is elected here.
                max_workers = min(32, (os.cpu_count() or 1) + 4)

                with ThreadPoolExecutor(max_workers=max_workers) as executor:
                    self.logger.debug("Max Workers: {}".format(max_workers))
                    while (
                        not exit_event.wait(timeout=loop_intervals[index])
                        and remaining_polls
                    ):
                        self.heartbeat()

                        if self.agent_process(executor):
                            index = 0
                        elif index < max(loop_intervals.keys()):
                            index += 1

                        remaining_polls -= 1

                        self.logger.debug(
                            "Next query for flow runs in {} seconds".format(
                                loop_intervals[index]
                            )
                        )
        finally:
            self.cleanup()

    def setup(self) -> None:
        self.agent_connect()

        if self.agent_address:
            parsed = urlparse(self.agent_address)
            if not parsed.port:
                raise ValueError("Must specify port in agent address")
            port = cast(int, parsed.port)
            hostname = parsed.hostname or ""
            app = web.Application([("/api/health", HealthHandler)])

            def run() -> None:
                self.logger.debug(
                    f"Agent API server listening on port {self.agent_address}"
                )
                self._api_server = app.listen(port, address=hostname)  # type: ignore
                self._api_server_loop = IOLoop.current()
                self._api_server_loop.start()  # type: ignore

            self._api_server_thread = threading.Thread(
                name="api-server", target=run, daemon=True
            )
            self._api_server_thread.start()

    def cleanup(self) -> None:
        self.on_shutdown()

        if self._api_server is not None:
            self.logger.debug("Stopping agent API server")
            self._api_server.stop()

        if self._api_server_loop is not None:
            self.logger.debug("Stopping agent API server loop")

            def stop_server() -> None:
                try:
                    loop = cast(IOLoop, self._api_server_loop)
                    loop.stop()
                except Exception:
                    pass

            self._api_server_loop.add_callback(stop_server)

        if self._api_server_thread is not None:
            self.logger.debug("Joining agent API threads")
            # Give the server a small period to shutdown nicely, otherwise it
            # will terminate on exit anyway since it's a daemon thread.
            self._api_server_thread.join(timeout=1)

    def on_shutdown(self) -> None:
        """
        Invoked when the event loop is exiting and the agent is shutting down. Intended
        as a hook for child classes to optionally implement.
        """

    def agent_connect(self) -> None:
        """
        Verify agent connection to Prefect API by querying
        """
        print(ascii_name)
        self.logger.info(
            "Starting {} with labels {}".format(type(self).__name__, self.labels)
        )
        self.logger.info(
            "Agent documentation can be found at https://docs.prefect.io/orchestration/"
        )

        self.logger.info(
            "Agent connecting to the Prefect API at {}".format(config.cloud.api)
        )
        try:
            self.client.graphql(query="query { hello }")
        except Exception as exc:
            self.logger.error(
                "There was an error connecting to {}".format(config.cloud.api)
            )
            self.logger.error(exc)

        self.logger.info("Waiting for flow runs...")

    def deploy_and_update_flow_run(self, flow_run: "GraphQLResult") -> None:
        """
        Deploy a flow run and update Cloud with the resulting deployment info.
        If any errors occur when submitting the flow run, capture the error and log to Cloud.

        Args:
            - flow_run (GraphQLResult): The specific flow run to deploy
        """
        # Deploy flow run and mark failed if any deployment error
        try:
            self.update_state(flow_run)
            deployment_info = self.deploy_flow(flow_run)
            if getattr(flow_run, "id", None):
                self.client.write_run_logs(
                    [
                        dict(
                            flow_run_id=getattr(flow_run, "id"),  # type: ignore
                            name=self.name,
                            message="Submitted for execution: {}".format(
                                deployment_info
                            ),
                            level="INFO",
                        )
                    ]
                )
        except Exception as exc:
            ## if the state update failed, we don't want to follow up with another state update
            if "State update failed" in str(exc):
                self.logger.debug("Updating Flow Run state failed: {}".format(str(exc)))
                return
            self.logger.error(
                "Logging platform error for flow run {}".format(
                    getattr(flow_run, "id", "UNKNOWN")  # type: ignore
                )
            )
            if getattr(flow_run, "id", None):
                self.client.write_run_logs(
                    [
                        dict(
                            flow_run_id=getattr(flow_run, "id"),  # type: ignore
                            name=self.name,
                            message=str(exc),
                            level="ERROR",
                        )
                    ]
                )
            self.mark_failed(flow_run=flow_run, exc=exc)

    def on_flow_run_deploy_attempt(self, fut: "Future", flow_run_id: str) -> None:
        """
        Indicates that a flow run deployment has been deployed (successfully or otherwise).
        This is intended to be a future callback hook, called in the agent's main thread
        when the background thread has completed the deploy_and_update_flow_run() call, either
        successfully, in error, or cancelled. In all cases the agent should be open to
        attempting to deploy the flow run if the flow run id is still in the Cloud run queue.

        Args:
            - fut (Future): a callback requirement, the future which has completed or been cancelled.
            - flow_run_id (str): the id of the flow run that the future represents.
        """
        self.submitting_flow_runs.remove(flow_run_id)
        self.logger.debug("Completed flow run submission (id: {})".format(flow_run_id))

    def agent_process(self, executor: "ThreadPoolExecutor") -> bool:
        """
        Full process for finding flow runs, updating states, and deploying.

        Args:
            - executor (ThreadPoolExecutor): the interface to submit flow deployments in background threads

        Returns:
            - bool: whether or not flow runs were found
        """
        flow_runs = None
        try:
            flow_runs = self.query_flow_runs()

            if flow_runs:
                self.logger.info(
                    "Found {} flow run(s) to submit for execution.".format(
                        len(flow_runs)
                    )
                )

            for flow_run in flow_runs:
                fut = executor.submit(self.deploy_and_update_flow_run, flow_run)
                self.submitting_flow_runs.add(flow_run.id)
                fut.add_done_callback(
                    functools.partial(
                        self.on_flow_run_deploy_attempt, flow_run_id=flow_run.id
                    )
                )

        except Exception as exc:
            self.logger.error(exc)

        return bool(flow_runs)

    def query_flow_runs(self) -> list:
        """
        Query Prefect Cloud for flow runs which need to be deployed and executed

        Returns:
            - list: A list of GraphQLResult flow run objects
        """
        self.logger.debug("Querying for flow runs")
        # keep a copy of what was curringly running before the query (future callbacks may be updating this set)
        currently_submitting_flow_runs = self.submitting_flow_runs.copy()

        # Get scheduled flow runs from queue
        mutation = {
            "mutation($input: get_runs_in_queue_input!)": {
                "get_runs_in_queue(input: $input)": {"flow_run_ids"}
            }
        }

        now = pendulum.now("UTC")
        result = self.client.graphql(
            mutation,
            variables={
                "input": {"before": now.isoformat(), "labels": list(self.labels),}
            },
        )

        # we queried all of the available flow runs, however, some may have already been pulled
        # by this agent and are in the process of being submitted in the background. We do not
        # want to act on these "duplicate" flow runs until we've been assured that the background
        # thread has attempted to submit the work (successful or otherwise).
        flow_run_ids = set(result.data.get_runs_in_queue.flow_run_ids)  # type: ignore

        if flow_run_ids:
            msg = "Found flow runs {}".format(
                result.data.get_runs_in_queue.flow_run_ids
            )
        else:
            msg = "No flow runs found"

        already_submitting = flow_run_ids & currently_submitting_flow_runs
        target_flow_run_ids = flow_run_ids - already_submitting

        if already_submitting:
            msg += " ({} already submitting: {})".format(
                len(already_submitting), list(already_submitting)
            )

        self.logger.debug(msg)

        # Query metadata fow flow runs found in queue
        query = {
            "query": {
                with_args(
                    "flow_run",
                    {
                        # match flow runs in the flow_run_ids list
                        "where": {
                            "id": {"_in": list(target_flow_run_ids)},
                            "_or": [
                                # who are EITHER scheduled...
                                {"state": {"_eq": "Scheduled"}},
                                # OR running with task runs scheduled to start more than 3 seconds ago
                                {
                                    "state": {"_eq": "Running"},
                                    "task_runs": {
                                        "state_start_time": {
                                            "_lte": str(now.subtract(seconds=3))  # type: ignore
                                        }
                                    },
                                },
                            ],
                        }
                    },
                ): {
                    "id": True,
                    "version": True,
                    "state": True,
                    "serialized_state": True,
                    "parameters": True,
                    "flow": {"id", "name", "environment", "storage", "version"},
                    with_args(
                        "task_runs",
                        {
                            "where": {
                                "state_start_time": {
                                    "_lte": str(now.subtract(seconds=3))  # type: ignore
                                }
                            }
                        },
                    ): {"id", "version", "task_id", "serialized_state"},
                }
            }
        }

        if target_flow_run_ids:
            self.logger.debug("Querying flow run metadata")
            result = self.client.graphql(query)
            return result.data.flow_run  # type: ignore
        else:
            return []

    def update_state(self, flow_run: GraphQLResult) -> None:
        """
        After a flow run is grabbed this function sets the state to Submitted so it
        won't be picked up by any other processes

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object
        """
        self.logger.debug(
            "Updating states for flow run {}".format(flow_run.id)  # type: ignore
        )

        # Set flow run state to `Submitted` if it is currently `Scheduled`
        if state.StateSchema().load(flow_run.serialized_state).is_scheduled():

            self.logger.debug(
                "Flow run {} is in a Scheduled state, updating to Submitted".format(
                    flow_run.id  # type: ignore
                )
            )
            self.client.set_flow_run_state(
                flow_run_id=flow_run.id,
                version=flow_run.version,
                state=Submitted(
                    message="Submitted for execution",
                    state=state.StateSchema().load(flow_run.serialized_state),
                ),
            )

        # Set task run states to `Submitted` if they are currently `Scheduled`
        for task_run in flow_run.task_runs:
            if state.StateSchema().load(task_run.serialized_state).is_scheduled():

                self.logger.debug(
                    "Task run {} is in a Scheduled state, updating to Submitted".format(
                        task_run.id  # type: ignore
                    )
                )
                self.client.set_task_run_state(
                    task_run_id=task_run.id,
                    version=task_run.version,
                    state=Submitted(
                        message="Submitted for execution.",
                        state=state.StateSchema().load(task_run.serialized_state),
                    ),
                )

    def mark_failed(self, flow_run: GraphQLResult, exc: Exception) -> None:
        """
        Mark a flow run as `Failed`

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object
            - exc (Exception): An exception that was raised to use as the `Failed`
                message
        """
        self.client.set_flow_run_state(
            flow_run_id=flow_run.id,
            version=flow_run.version,
            state=Failed(message=str(exc)),
        )
        self.logger.error("Error while deploying flow: {}".format(repr(exc)))

    def deploy_flow(self, flow_run: GraphQLResult) -> str:
        """
        Meant to be overridden by a platform specific deployment option

        Args:
            - flow_run (GraphQLResult): A GraphQLResult flow run object

        Returns:
            - str: Information about the deployment

        Raises:
            - ValueError: if deployment attempted on unsupported Storage type
        """
        raise NotImplementedError()

    def heartbeat(self) -> None:
        """
        Meant to be overridden by a platform specific heartbeat option
        """


if __name__ == "__main__":
    Agent().start()



from typing import Any

import prefect
from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

try:
    from jinja2 import Template
except ImportError:
    raise ImportError(
        'Using `prefect.tasks.templates.jinja2` requires Prefect to be installed with the "templates" extra.'
    )


class JinjaTemplate(Task):
    """
    This task contains a Jinja template which is formatted with the results of any
    upstream tasks and returned.

    Variables from `prefect.context` will also be used for rendering.

    Args:
        - template (str, optional): the optional _default_ template string to render at runtime;
            can also be provided as a keyword to `run`, which takes precendence over this default.
        - **kwargs (optional): additional keyword arguments to pass to the
            standard Task constructor

    Example:

    ```python
    from prefect import Flow
    from prefect.tasks.templates import JinjaTemplate


    message = '''
    Hi {{name}}!  Welcome to Prefect.  Today is {{today}}.
    '''

    msg_task = JinjaTemplate(name="message body", template=message)

    with Flow("string-template") as flow:
            output = msg_task(name="Marvin")

    flow_state = flow.run()

    print(flow_state.result[output].result)
    # Hi Marvin!  Welcome to Prefect.  Today is 2019-08-28.
    ```
    """

    def __init__(self, template: str = None, **kwargs: Any):
        self.template = template or ""
        super().__init__(**kwargs)

    @defaults_from_attrs("template")
    def run(self, template: str = None, **format_kwargs: Any) -> str:
        """
        Formats the Jinja Template with the provided kwargs.

        Args:
            - template (str, optional): the template string to render; if not
                provided, `self.template` will be used
            - **format_kwargs (optional): keyword arguments to use for
                rendering; note that variables from `prefect.context` will also be used

        Returns:
            - str: the rendered string
        """
        template = Template(template)
        with prefect.context(**format_kwargs) as data:
            return template.render(**data)



import collections
from collections.abc import MutableMapping
from typing import Any, Generator, Iterable, Iterator, Union, cast

from box import Box

DictLike = Union[dict, "DotDict"]


def flatten_seq(seq: Iterable) -> Generator:
    """
    Generator that returns a flattened list from a possibly nested list-of-lists
    (or any sequence type).

    Example:
        ```python
        flatten_seq([1, 2, [3, 4], 5, [6, [7]]])
        >>> [1, 2, 3, 4, 5, 6, 7]
        ```

    Args:
        - seq (Iterable): the sequence to flatten

    Returns:
        - generator: a generator that yields the flattened sequence
    """
    for item in seq:
        if isinstance(item, collections.abc.Iterable) and not isinstance(
            item, (str, bytes)
        ):
            yield from flatten_seq(item)
        else:
            yield item


class DotDict(MutableMapping):
    """
    A `dict` that also supports attribute ("dot") access. Think of this as an extension
    to the standard python `dict` object.  **Note**: while any hashable object can be added to
    a `DotDict`, _only_ valid Python identifiers can be accessed with the dot syntax; this excludes
    strings which begin in numbers, special characters, or double underscores.

    Args:
        - init_dict (dict, optional): dictionary to initialize the `DotDict`
        with
        - **kwargs (optional): key, value pairs with which to initialize the
        `DotDict`

    Example:
        ```python
        dotdict = DotDict({'a': 34}, b=56, c=set())
        dotdict.a # 34
        dotdict['b'] # 56
        dotdict.c # set()
        ```
    """

    def __init__(self, init_dict: DictLike = None, **kwargs: Any):
        # a DotDict could have a key that shadows `update`
        if init_dict:
            super().update(init_dict)
        super().update(kwargs)

    def get(self, key: str, default: Any = None) -> Any:
        """
        This method is defined for MyPy, which otherwise tries to type
        the inherited `.get()` method incorrectly.

        Args:
            - key (str): the key to retrieve
            - default (Any): a default value to return if the key is not found

        Returns:
            - Any: the value of the key, or the default value if the key is not found
        """
        return super().get(key, default)

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]  # __dict__ expects string keys

    def __setitem__(self, key: str, value: Any) -> None:
        self.__dict__[key] = value

    def __setattr__(self, attr: str, value: Any) -> None:
        self[attr] = value

    def __iter__(self) -> Iterator[str]:
        return iter(self.__dict__.keys())

    def __delitem__(self, key: str) -> None:
        del self.__dict__[key]

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        if len(self) > 0:
            return "<{}: {}>".format(
                type(self).__name__, ", ".join(sorted(repr(k) for k in self.keys()))
            )
        else:
            return "<{}>".format(type(self).__name__)

    def copy(self) -> "DotDict":
        """Creates and returns a shallow copy of the current DotDict"""
        return type(self)(self.__dict__.copy())

    def to_dict(self) -> dict:
        """
        Converts current `DotDict` (and any `DotDict`s contained within)
        to an appropriate nested dictionary.
        """
        # mypy cast
        return cast(dict, as_nested_dict(self, dct_class=dict))


def merge_dicts(d1: DictLike, d2: DictLike) -> DictLike:
    """
    Updates `d1` from `d2` by replacing each `(k, v1)` pair in `d1` with the
    corresponding `(k, v2)` pair in `d2`.

    If the value of each pair is itself a dict, then the value is updated
    recursively.

    Args:
        - d1 (MutableMapping): A dictionary to be replaced
        - d2 (MutableMapping): A dictionary used for replacement

    Returns:
        - A `MutableMapping` with the two dictionary contents merged
    """

    new_dict = d1.copy()

    for k, v in d2.items():
        if isinstance(new_dict.get(k), MutableMapping) and isinstance(
            v, MutableMapping
        ):
            new_dict[k] = merge_dicts(new_dict[k], d2[k])
        else:
            new_dict[k] = d2[k]
    return new_dict


def as_nested_dict(
    obj: Union[DictLike, Iterable[DictLike]], dct_class: type = DotDict
) -> Union[DictLike, Iterable[DictLike]]:
    """
    Given a obj formatted as a dictionary, transforms it (and any nested dictionaries)
    into the provided dct_class

    Args:
        - obj (Any): An object that is formatted as a `dict`
        - dct_class (type): the `dict` class to use (defaults to DotDict)

    Returns:
        - A `dict_class` representation of the object passed in
    ```
    """
    if isinstance(obj, (list, tuple, set)):
        return type(obj)([as_nested_dict(d, dct_class) for d in obj])

    # calling as_nested_dict on `Box` objects pulls out their "private" keys due to our recursion
    # into `__dict__` if it exists. We can special-case Box and just convert it to dict this way,
    # which automatically handles recursion.
    elif isinstance(obj, Box):
        return dict(obj)
    elif isinstance(obj, (dict, DotDict)):
        # DotDicts could have keys that shadow `update` and `items`, so we
        # take care to avoid accessing those keys here
        return dct_class(
            {
                k: as_nested_dict(v, dct_class)
                for k, v in getattr(obj, "__dict__", obj).items()
            }
        )
    return obj


class CompoundKey(tuple):
    pass


def dict_to_flatdict(dct: dict, parent: CompoundKey = None) -> dict:
    """Converts a (nested) dictionary to a flattened representation.

    Each key of the flat dict will be a CompoundKey tuple containing the "chain of keys"
    for the corresponding value.

    Args:
        - dct (dict): The dictionary to flatten
        - parent (CompoundKey, optional): Defaults to `None`. The parent key
        (you shouldn't need to set this)

    Returns:
        - dict: A flattened dict
    """

    items = []  # type: list
    parent = parent or CompoundKey()
    for k, v in dct.items():
        k_parent = CompoundKey(parent + (k,))
        if isinstance(v, dict):
            items.extend(dict_to_flatdict(v, parent=k_parent).items())
        else:
            items.append((k_parent, v))
    return dict(items)


def flatdict_to_dict(dct: dict, dct_class: type = None) -> MutableMapping:
    """Converts a flattened dictionary back to a nested dictionary.

    Args:
        - dct (dict): The dictionary to be nested. Each key should be a
        `CompoundKey`, as generated by `dict_to_flatdict()`
        - dct_class (type, optional): the type of the result; defaults to `dict`

    Returns:
        - MutableMapping: A `MutableMapping` used to represent a nested dictionary
    """

    result = (dct_class or dict)()  # type: MutableMapping
    for k, v in dct.items():
        if isinstance(k, CompoundKey):
            current_dict = result
            for ki in k[:-1]:
                current_dict = current_dict.setdefault(  # type: ignore
                    ki, (dct_class or dict)()
                )
            current_dict[k[-1]] = v
        else:
            result[k] = v

    return result



from typing import Any, Callable

from prefect.engine.executors.base import Executor


class LocalExecutor(Executor):
    """
    An executor that runs all functions synchronously and immediately in
    the main thread.  To be used mainly for debugging purposes.
    """

    def submit(
        self, fn: Callable, *args: Any, extra_context: dict = None, **kwargs: Any
    ) -> Any:
        """
        Submit a function to the executor for execution. Returns the result of the computation.

        Args:
            - fn (Callable): function that is being submitted for execution
            - *args (Any): arguments to be passed to `fn`
            - extra_context (dict, optional): an optional dictionary with extra information about the submitted task
            - **kwargs (Any): keyword arguments to be passed to `fn`

        Returns:
            - Any: the result of `fn(*args, **kwargs)`
        """
        return fn(*args, **kwargs)

    def wait(self, futures: Any) -> Any:
        """
        Returns the results of the provided futures.

        Args:
            - futures (Any): objects to wait on

        Returns:
            - Any: whatever `futures` were provided
        """
        return futures



import json
from base64 import b64encode

from prefect import Task
from prefect.utilities.aws import get_boto_client
from prefect.utilities.tasks import defaults_from_attrs


class LambdaCreate(Task):
    """
    Task for creating a Lambda function.

    Args:
        - function_name (str): name of the Lambda function to create
        - runtime (str): the identifier of the function's runtime
        - role (str): the Amazon Resource Name of the function's execution role
        - handler (str): the name of the method within your code that Lambda calls
            to execute your function
        - zip_file (str): path to zip file containing code for Lambda function,
            either zip_file or (bucket and bucket_key) must be passed
        - bucket (str): an S3 bucket in the same AWS region as your function
        - bucket_key (str): the Amazon S3 key of the deployment package
        - object_version (str, optional): for versioned S3 objects, the version of the
            deployment package to use
        - description (str, optional): description of Lambda function
        - function_timeout (int, optional): Lambda function timeout in seconds, default is 3 seconds
        - memorysize (int, optional): amount of memory that Lambda function has
            access to in MB, must be a multiple of 64 MB, default is 128
        - publish (bool, optional): set to True to publish the first version of the
            function during creation, defaults to True
        - subnet_ids (List[str], optional): list of subnet ids for vpc
            configuration
        - security_group_ids (List[str], optional): list of security
            group ideas for vpc configuration
        - dead_letter_config (dict, optional): a dead letter queue configuration that
            specifies the queue or topic where Lambda sends asynchronous events
            when they fail processing
        - environment_variables (dict, optional): key-value pairs of environment
            variables to pass to the Lambda function
        - kms_key_arn (str, optional): the ARN of the AWS key management service used
            to encrypt your function's environment variables, if not provided, AWS
            Lambda uses a default service key
        - function_tags (dict, optional): a list of tags to apply to the function, string
            to string map
        - tracing_config (str, optional): set to Active to samle and trace a
            subset of incoming requests with Amazon X-Ray
        - layers (List[str], optional): a list of function layers to add to
            the function's execution environment, specify each layer by its ARN
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        function_name: str,
        runtime: str,
        role: str,
        handler: str,
        zip_file: str = None,
        bucket: str = "",
        bucket_key: str = "",
        object_version: str = None,
        description: str = "",
        function_timeout: int = 3,
        memorysize: int = 128,
        publish: bool = True,
        subnet_ids: list = None,
        security_group_ids: list = None,
        dead_letter_config: dict = None,
        environment_variables: dict = None,
        kms_key_arn: str = "",
        function_tags: dict = None,
        tracing_config: str = "PassThrough",
        layers: list = None,
        **kwargs
    ):
        self.function_name = function_name
        self.runtime = runtime
        self.role = role
        self.handler = handler

        ## if zip file is provided, pass this to boto3 create, otherwise pass s3 object
        if zip_file:
            self.code = {"ZipFile": open(zip_file, "rb").read()}
        else:
            self.code = {"S3Bucket": bucket, "S3Key": bucket_key}
            if object_version:
                self.code["S3ObjectVersion"] = object_version

        self.description = description
        self.function_timeout = function_timeout
        self.memorysize = memorysize
        self.publish = publish

        self.vpc_config = {}
        if subnet_ids:
            self.vpc_config["SubnetIds"] = subnet_ids
        if security_group_ids:
            self.vpc_config["SecurityGroupIds"] = security_group_ids

        self.dead_letter_config = dead_letter_config
        self.environment_variables = environment_variables
        self.kms_key_arn = kms_key_arn
        self.function_tags = function_tags
        self.tracing_config = tracing_config
        self.layers = layers

        super().__init__(**kwargs)

    def run(self, credentials: dict = None):
        """
        Task run method. Creates Lambda function.

        Args:
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.

        Returns:
            - json: response from AWS CreateFunction endpoint
        """

        lambda_client = get_boto_client("lambda", credentials=credentials)

        ## create lambda function
        response = lambda_client.create_function(
            FunctionName=self.function_name,
            Runtime=self.runtime,
            Role=self.role,
            Handler=self.handler,
            Code=self.code,
            Description=self.description,
            Timeout=self.function_timeout,
            MemorySize=self.memorysize,
            Publish=self.publish,
            VpcConfig=self.vpc_config,
            DeadLetterConfig=self.dead_letter_config or {},
            Environment={"Variables": self.environment_variables or {}},
            KMSKeyArn=self.kms_key_arn,
            TracingConfig={"Mode": self.tracing_config},
            Tags=self.function_tags or {},
            Layers=self.layers or [],
        )

        return response


class LambdaDelete(Task):
    """
    Task for deleting a Lambda function.

    Args:
        - function_name (str): name of the Lambda function to delete
        - qualifier (str, optional): specify a version to delete, if not
            provided, the function will be deleted entirely
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(self, function_name: str, qualifier: str = "", **kwargs):
        self.function_name = function_name
        self.qualifier = qualifier
        super().__init__(**kwargs)

    def run(self, credentials: dict = None):
        """
        Task run method. Deletes Lambda function.

        Args:
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.

        Returns:
            - dict: response from AWS DeleteFunction endpoint
        """

        lambda_client = get_boto_client("lambda", credentials=credentials)

        ## delete function, depending on if qualifier provided
        if len(self.qualifier) > 0:
            response = lambda_client.delete_function(
                FunctionName=self.function_name, Qualifier=self.qualifier
            )
            return response

        response = lambda_client.delete_function(FunctionName=self.function_name)
        return response


class LambdaInvoke(Task):
    """
    Task to invoke a Lambda function.

    Args:
        - function_name (str): the name of the Lambda funciton to invoke
        - invocation_type (str, optional): the invocation type of Lambda
            function, default is RequestResponse other options include
            Event and DryRun
        - log_type (str, optional): set to 'Tail' to include the execution
            log in the response
        - client_context (dict, optional): data to pass to the function in the
            context object, dict object will be transformed into base64 encoded
            json automatically
        - payload (bytes or seekable file-like object): the JSON provided to
            Lambda function as input
        - qualifier (str, optional): specify a version or alias to invoke a
            published version of the function, defaults to $LATEST
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        function_name: str,
        invocation_type: str = "RequestResponse",
        log_type: str = "None",
        client_context: dict = None,
        payload: str = json.dumps(None),
        qualifier: str = "$LATEST",
        **kwargs
    ):
        self.function_name = function_name
        self.invocation_type = invocation_type
        self.log_type = log_type

        ## encode input dictionary as base64 json
        self.client_context = self._encode_lambda_context(**(client_context or {}))

        self.payload = payload
        self.qualifier = qualifier
        super().__init__(**kwargs)

    def _encode_lambda_context(self, custom=None, env=None, client=None):

        """
        Utility function for encoding Lambda context

        Args:
            - custom (dict, optional): key-value  pairs to pass to custom context
            - env (dict, optional): key-value pairs to pass to environment context
            - client (dict, optional): key-value pairs to pass to client context

        Returns:
            - json: base64 encoded json object
        """

        client_context = dict(custom=custom, env=env, client=client)
        json_context = json.dumps(client_context).encode("utf-8")
        return b64encode(json_context).decode("utf-8")

    @defaults_from_attrs("function_name", "payload")
    def run(
        self, function_name: str = None, payload: str = None, credentials: dict = None,
    ):
        """
        Task run method. Invokes Lambda function.

        Args:
            - function_name (str): the name of the Lambda funciton to invoke
            - payload (bytes or seekable file-like object): the JSON provided to
                Lambda function as input
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.

        Returns:
            - dict : response from AWS Invoke endpoint
        """

        lambda_client = get_boto_client("lambda", credentials=credentials)

        ## invoke lambda function
        response = lambda_client.invoke(
            FunctionName=function_name,
            InvocationType=self.invocation_type,
            LogType=self.log_type,
            ClientContext=self.client_context,
            Payload=payload,
            Qualifier=self.qualifier,
        )

        return response


class LambdaList(Task):
    """
    Task to list Lambda functions.

    Args:
        - master_region (str, optional): for Lambda@Edge functions, the AWS
            region of the master function
        - function_version (str, optional): the version of a function,
            default is 'ALL'
        - marker (str, optional): specify the pagination token that's returned
            by a previous request to retreive the next page of results
        - max_items (int, optional): specify a value between 1 and 50 to limit
            the number of functions in the response
        - **kwargs (dict, optional): additional keyword arguments to pass to the
            Task constructor
    """

    def __init__(
        self,
        master_region: str = "ALL",
        function_version: str = "ALL",
        marker: str = None,
        max_items: int = 50,
        **kwargs
    ):
        self.master_region = master_region
        self.function_version = function_version
        self.marker = marker
        self.max_items = max_items
        super().__init__(**kwargs)

    def run(self, credentials: dict = None):
        """
        Task fun method. Lists all Lambda functions.

        Args:
            - credentials (dict, optional): your AWS credentials passed from an upstream
                Secret task; this Secret must be a JSON string
                with two keys: `ACCESS_KEY` and `SECRET_ACCESS_KEY` which will be
                passed directly to `boto3`.  If not provided here or in context, `boto3`
                will fall back on standard AWS rules for authentication.

        Returns:
            - dict : a list of Lambda functions from AWS ListFunctions endpoint
        """

        lambda_client = get_boto_client("lambda", credentials=credentials)

        ## list functions, optionally passing in marker if not None
        if self.marker:
            response = lambda_client.list_functions(
                MasterRegion=self.master_region,
                FunctionVersion=self.function_version,
                Marker=self.marker,
                MaxItems=self.max_items,
            )
            return response

        response = lambda_client.list_functions(
            MasterRegion=self.master_region,
            FunctionVersion=self.function_version,
            MaxItems=self.max_items,
        )

        return response
